import pytest

import hal_impl
from . import fake_time, pyfrc_fake_hooks

from .controller import TestController


class ThreadStillRunningError(Exception):
    pass


class PyFrcPlugin:
    """
    Pytest plugin. Each documented member function name can be an argument
    to your test functions, and the data that these functions return will
    be passed to your test function.
    """

    def __init__(self, robot_class, robot_file, robot_path):
        self.robot_class = robot_class

        self._robot_file = robot_file
        self._robot_path = robot_path

        # Setup fake time
        self._fake_time = fake_time.FakeTime()

        # Setup control instance
        self._control = None

        self._started = False

        # Setup the hal hooks so we can control time
        # -> The hook doesn't have any state, so we initialize it only once
        hal_impl.functions.hooks = pyfrc_fake_hooks.PyFrcFakeHooks(self._fake_time)

    def pytest_runtest_setup(self):
        # This function needs to do the same things that RobotBase.main does,
        # plus some extra things needed for testing

        #
        # Initialization order is important, because there are a bunch
        # of inter-related things happening here.
        #
        # - Initialize networktables first since it has no dependencies
        # - Initialize the fake time, because the HAL uses it
        #   - This initializes the driver station, but it doesn't need HAL yet
        # - Reset HAL data
        # - Initialize RobotBase
        #

        import networktables

        if hasattr(networktables, "NetworkTables"):
            if hasattr(networktables.NetworkTables, "setTestMode"):
                networktables.NetworkTables.setTestMode()
            else:
                networktables.NetworkTables.startTestMode()
        else:
            networktables.NetworkTable.setTestMode()

        self._fake_time.initialize()
        self._test_controller = TestController(self._fake_time)

        hal_impl.functions.reset_hal()

        import wpilib

        hwcfg = getattr(wpilib.RobotBase, "initializeHardwareConfiguration", None)
        if hwcfg:
            hwcfg()

        self._test_controller._robot = self.robot_class()

        # TODO: Remove after 2016
        getattr(self._test_controller._robot, "prestart", lambda: True)()

        assert hasattr(
            self._test_controller._robot, "_RobotBase__initialized"
        ), "If your robot class has an __init__ function, it must call super().__init__()!"

        self._started = True

    def pytest_runtest_teardown(self, nextitem):

        started = self._started
        self._started = False

        # Let any child threads run in realtime to allow cancelling if it
        # has been implemented.
        self._fake_time.teardown()
        self._test_controller = None

        # If the unit test never started, then the rest may hang. Bail
        # out now instead, and choose to have more errors later.
        if not started:
            return

        import wpilib._impl.utils

        wpilib._impl.utils.reset_wpilib()

        import networktables

        if hasattr(networktables, "NetworkTables"):
            networktables.NetworkTables.shutdown()
        else:
            networktables.NetworkTable._staticProvider.close()
            networktables.NetworkTable._staticProvider = None

        if not self._fake_time.children_stopped():
            raise ThreadStillRunningError(
                "Make sure spawning class has free() "
                "method to stop thread and is registered with "
                "Resource.add_global_resource()."
            )

    #
    # Fixtures
    #
    # Each one of these can be arguments to your test, and the result of the
    # corresponding function will be passed to your test as that argument.
    #

    @pytest.fixture()
    def control(self):
        """
        A fixture that provides control over the robot

        :rtype: :class:`.TestController`
        """
        return self._test_controller

    @pytest.fixture()
    def fake_time(self):
        """
        A fixture that gives you control over the time your robot is using

        :rtype:   :class:`.FakeTime`
        """
        return self._fake_time

    @pytest.fixture()
    def hal_data(self):
        """
        Provides access to a dict with all the device data about the robot

        .. seealso:: For a listing of what the dict contains and some documentation, see
                     https://github.com/robotpy/robotpy-wpilib/blob/master/hal-sim/hal_impl/data.py
        """
        return hal_impl.data.hal_data

    @pytest.fixture()
    def robot(self):
        """Your robot instance"""
        return self._test_controller._robot

    @pytest.fixture()
    def robot_file(self):
        """The absolute filename your robot code is started from"""
        return self._robot_file

    @pytest.fixture()
    def robot_path(self):
        """The absolute directory that your robot code is located at"""
        return self._robot_path

    @pytest.fixture()
    def wpilib(self):
        """The :mod:`wpilib` module. Provided for backwards compatibility"""
        import wpilib

        return wpilib

    #
    # Internal use only because newer versions of pytest
    # don't allow calling fixtures directly
    #

    def get_control(self):
        return self._test_controller

    def get_fake_time(self):
        return self._fake_time

    def get_robot(self):
        return self._test_controller._robot



"""
    pyfrc uses the pint library in some places for representing physical
    quantities to allow users to specify the physical parameters of their robot
    in a natural and non-ambiguous way. For example, to represent 5 feet::
    
        from pyfrc.physics.units import units
        
        five_feet = 5 * units.feet
        
    Unfortunately, actually using the quantities is a huge performance hit, so
    we don't use them to perform actual physics computations. Instead, pyfrc uses
    them to convert to known units, then performs computations using
    the magnitude of the quantity.
    
    pyfrc defines the following custom units:
    
    * ``counts_per_minute`` or ``cpm``: Counts per minute, which should be used
      instead of pint's predefined ``rpm`` (because it is rad/s). Used to
      represent motor free speed
    * ``N_m``: Shorthand for N-m or newton-meter. Used for motor torque.
    
    * ``tm_ka``: The kA value used in the tankmodel (uses imperial units)
    * ``tm_kv``: The kV value used in the tankmodel (uses imperial units)
    
    Refer to the `pint documentation <https://pint.readthedocs.io>`_ for more
    information on how to use pint.
"""


import pint

#: All units that pyfrc uses are defined in this global object
units = pint.UnitRegistry()

# Shorthand notation used to represent motor torque
units.define("N_m = newton * meter")

# Counts per minute. pint defines RPM, but it is in rad/s, which is not often
# what we want
units.define("counts_per_minute = count / minute = cpm")

# Special units used in the tankmodel
units.define("tm_kv = volt / (foot / second)")
units.define("tm_ka = volt / (foot / second ** 2)")

# Helper functions
class Helpers:
    ensure_mass = units.check("[mass]")(lambda u: u)
    ensure_length = units.check("[length]")(lambda u: u)

    ensure_acceleration = units.check("[length] / [time]**2")(lambda u: u)
    ensure_time = units.check("[time]")(lambda u: u)
    ensure_velocity = units.check("[length] / [time]")(lambda u: u)



import inspect

from hal_impl import mode_helpers
from hal_impl.data import hal_data

from .fake_time import TestEnded


class _PracticeMatch:

    autonomous_period = 15
    operator_period = 15

    def __init__(self, on_step, controller):
        self._on_step = on_step
        self._controller = controller

    def on_step(self, tm):
        """
        Called when a driver station packet would be delivered
        """

        if tm < 5:
            self._controller.set_autonomous(False)

        elif tm < 5 + self.autonomous_period:
            self._controller.set_autonomous(True)

        elif tm < 5 + self.autonomous_period + 1:
            self._controller.set_operator_control(False)

        elif tm < 5 + self.autonomous_period + 1 + self.operator_period:
            self._controller.set_operator_control(True)

        else:
            return False

        if self._on_step is not None:
            retval = self._on_step(tm)
            if retval is not None:
                return retval

        return True


class TestController:
    """
    This object is used to control the robot during unit tests. You
    do not need to create an instance of this, instead use the
    ``controller`` fixture.

    To set a game specific message for autonomous mode, just
    set the 'game_specific_message' property of this object, and it will
    be setup correctly upon transition into autonomous mode.
    """

    def __init__(self, fake_time_inst):
        self._practice = False
        self._test_running = False
        self._ds_cond = fake_time_inst.ds_cond
        self.game_specific_message = None

    def get_mode(self):
        """Returns the current mode that the robot is in

        :returns: 'autonomous', 'teleop', 'test', or 'disabled'
        """

        ctrl = hal_data["control"]
        if not ctrl["enabled"]:
            return "disabled"
        if ctrl["autonomous"]:
            return "autonomous"
        if ctrl["test"]:
            return "test"

        return "teleop"

    def set_practice_match(self):
        """Call this function to enable a practice match. Must only
        be called before :meth:`run_test` is called.
        """

        assert not self._test_running
        self._practice = True

    def set_autonomous(self, enabled=True):
        """Puts the robot in autonomous mode"""
        mode_helpers.set_mode(
            "auto", enabled, game_specific_message=self.game_specific_message
        )

    def set_operator_control(self, enabled=True):
        """Puts the robot in operator control mode"""
        mode_helpers.set_mode("teleop", enabled)

    def set_test_mode(self, enabled=True):
        """Puts the robot in test mode (the robot mode, not related to unit testing)"""
        mode_helpers.set_mode("test", enabled)

    def run_test(self, controller=None):
        """
        Call this to execute the robot code. Cannot be called more than once
        in a single test.

        If the controller argument is a class, it will be constructed and the
        instance will be returned.

        :param controller: This can either be a function that takes a single
                           argument, or a class that has an 'on_step' function.
                           If it is a class, an instance will be created. Either
                           the function or the on_step function will be called
                           with a single parameter, which is the the current
                           robot time. If None, an error will be signaled unless
                           :meth:`set_practice_match` has been called.
        """

        # Can't call this twice!
        assert not self._test_running

        on_step = None
        retval = None

        if controller is not None:

            if inspect.isclass(controller):
                retval = controller()
                on_step = retval.on_step

            elif callable(controller):
                on_step = controller

            else:
                raise ValueError("Invalid controller parameter")

            info = inspect.getfullargspec(on_step)
            if len(info.args) > 0 and info.args[0] == "self":
                info.args.remove("self")

            if len(info.args) + len(info.kwonlyargs) != 1:
                raise ValueError(
                    "%s must be a function that takes a single argument" % on_step
                )

        # Setup the time hooks
        if self._practice:
            pm = _PracticeMatch(on_step, self)
            on_step = pm.on_step

        if on_step is not None:
            self._ds_cond._on_step = on_step
        else:
            raise ValueError(
                "The controller parameter must not be None OR you must call control.set_practice_match"
            )

        self._test_running = True

        try:
            self._robot.startCompetition()
        except TestEnded:
            pass
        except:
            raise

        return retval



#
# Holds settings that can be checked
#

# Set to True if running in code coverage mode (only set when running tests)
# -> Since pyfrc 2014.6.0
coverage_mode = False

# Indicates how pyfrc was run: netsim/sim/upload/wipe, etc...
# -> Since pyfrc 2014.6.0
mode = None

# A dictionary of configuration information that is partially loaded
# from sim/config.json
# -> Since pyfrc 2018.1.0
config_obj = {}



# def pytest_runtest_setup():
#     import wpilib
#     import networktables

#     try:
#         networktables.NetworkTables.setTestMode()
#     except AttributeError:
#         networktables.NetworkTables.startTestMode()

#     wpilib.DriverStation._reset()



# import pytest
from pyfrc.physics import tankmodel, motor_cfgs
from pyfrc.physics.units import units
import math


def test_tankdrive_get_distance():
    """Test TankModel.get_distance() by driving in a quarter circle.
    If the turn angle is 90deg, then x should equal y. get_distance() runs in little steps,
    so multiple calls should be equivalent to a single call"""

    # just use the default robot
    tank = tankmodel.TankModel.theory(
        motor_cfgs.MOTOR_CFG_CIM,
        robot_mass=90 * units.lbs,
        gearing=10.71,
        nmotors=2,
        x_wheelbase=2.0 * units.feet,
        wheel_diameter=6 * units.inch,
    )
    # slight turn to the right
    l_motor = 1.0
    r_motor = -0.9

    # get the motors up to speed, so that subsequent calls produce the same result
    tank.calculate(l_motor, r_motor, 2.0)

    # figure out how much time is needed to turn 90 degrees
    angle = 0.0
    angle_target = -math.pi / 2.0  # 90 degrees
    total_time = 0.0
    tstep = 0.01

    while angle < angle_target:
        result = tank.calculate(l_motor, r_motor, tstep)
        angle += result.rotation().radians()
        total_time += tstep

    # print('time needed', total_time)

    # OK, now do it in a single call. x should equal y
    result = tank.calculate(l_motor, r_motor, total_time)
    # print(result[0], result[1], degrees(result[2]), result[0]/result[1])
    assert math.isclose(
        result.rotation().radians(), angle_target, abs_tol=2.0
    ), "Single call should produce a 90deg turn"
    assert math.isclose(
        result.translation().x, result.translation().y, rel_tol=0.01
    ), "For 90deg turn, x and y should be the same"
    return



import sys


def print_err(*args):
    print(*args, file=sys.stderr)


def yesno(prompt):
    """Returns True if user answers 'y'"""
    prompt += " [y/n]"
    a = ""
    while a not in ["y", "n"]:
        a = input(prompt).lower()

    return a == "y"



"""
    The primary purpose of these tests is to run through your code
    and make sure that it doesn't crash. If you actually want to test
    your code, you need to write your own custom tests to tease out
    the edge cases.
    
    To use these, add the following to a python file in your tests directory::
    
        from pyfrc.tests import *
    
"""

import math
import pytest

from .. import config

_gsms = config.config_obj["pyfrc"]["game_specific_messages"]


@pytest.mark.parametrize("gamedata", _gsms if _gsms else [""])
def test_autonomous(control, fake_time, robot, gamedata):
    """Runs autonomous mode by itself"""

    # run autonomous mode for 15 seconds
    control.game_specific_message = gamedata
    control.set_autonomous(enabled=True)
    control.run_test(lambda tm: tm < 15)

    # make sure autonomous mode ran for 15 seconds
    assert int(fake_time.get()) == 15


@pytest.mark.filterwarnings("ignore")
def test_disabled(control, fake_time, robot):
    """Runs disabled mode by itself"""

    # run disabled mode for 5 seconds
    control.set_autonomous(enabled=False)
    control.run_test(lambda tm: tm < 5)

    # make sure disabled mode ran for 5 seconds
    assert int(fake_time.get()) == 5


@pytest.mark.filterwarnings("ignore")
def test_operator_control(control, fake_time, robot):
    """Runs operator control mode by itself"""

    # run operator mode for 15 seconds
    control.set_operator_control(enabled=True)
    control.run_test(lambda tm: tm < 15)

    # make sure operator mode ran for 15 seconds
    assert int(fake_time.get()) == 15


@pytest.mark.filterwarnings("ignore")
def test_practice(control, fake_time, robot):
    """Runs through the entire span of a practice match"""

    class TestController:
        def __init__(self):
            self.mode = None

            self.disabled = 0
            self.autonomous = 0
            self.teleop = 0

        def on_step(self, tm):
            """
            Called on each simulation step. This runs through each mode,
            and asserts that the robot didn't spend too much time in any
            particular mode.

            If you get an assertion error here, it can mean a lot of
            different things, but typically it means you called Timer.delay()
            somewhere with a parameter greater than a few milliseconds..
            which is almost always a bad idea, and your robot will ignore
            your input for a few seconds.

            :param tm: The current robot time
            """

            mode = control.get_mode()
            if mode == self.mode:
                return

            if mode == "autonomous":
                self.autonomous += 1
                assert int(math.floor(fake_time.get())) == 5

            elif mode == "teleop":
                self.teleop += 1
                assert int(math.floor(fake_time.get())) == 21

            elif mode == "disabled":
                self.disabled += 1

                if self.disabled == 1:
                    # Time must be pretty close to zero at this point.
                    # If it's not, you called Timer.delay() somewhere you
                    # shouldn't have before the test started... you can
                    # try getting rid of the delay by only doing it when
                    # RobotBase.isSimulation() returns False
                    assert int(math.floor(fake_time.get())) == 0
                else:
                    assert int(math.floor(fake_time.get())) == 20
            else:
                assert False, "Internal error!"

            self.mode = mode

    if _gsms:
        control.game_specific_message = _gsms[0]

    control.set_practice_match()
    tc = control.run_test(TestController)

    assert int(math.floor(fake_time.get())) == 36

    # If an error occurs here, for some reason a mode got called too many times
    assert tc.disabled == 2
    assert tc.autonomous == 1
    assert tc.teleop == 1



"""
    These generic test modules can be applied to :class:`wpilib.iterativerobot.IterativeRobot`
    and :class:`wpilib.samplerobot.SampleRobot` based robots.
"""

# import basic tests
from .basic import test_autonomous, test_disabled, test_operator_control, test_practice

# Other test types
from .fuzz_test import test_fuzz



"""
    .. versionadded:: 2018.4.0

    .. note:: The equations used in our :class:`TankModel` is derived from
              `Noah Gleason and Eli Barnett's motor characterization whitepaper
              <https://www.chiefdelphi.com/media/papers/3402>`_. It is
              recommended that users of this model read the paper so they can
              more fully understand how this works.
              
              In the interest of making progress, this API may receive
              backwards-incompatible changes before the start of the 2019
              FRC season.
"""

import math
import typing

from wpilib.geometry import Transform2d

from .motor_cfgs import MotorModelConfig
from .units import units, Helpers

import logging

logger = logging.getLogger("pyfrc.physics")

# default parameters for a kitbot
_bumper_length = 3.25 * units.inch

_kitbot_wheelbase = 21.0 * units.inch
_kitbot_width = _kitbot_wheelbase + _bumper_length * 2
_kitbot_length = 30.0 * units.inch + _bumper_length * 2

_inertia_units = (units.foot ** 2) * units.pound
_bm_units = units.foot * units.pound


class MotorModel:
    """
    Motor model used by the :class:`TankModel`. You should not need to create
    this object if you're using the :class:`TankModel` class.
    """

    @units.wraps(None, (None, None, "tm_kv", "tm_ka", "volts"))
    def __init__(
        self,
        motor_config: MotorModelConfig,
        kv: units.tm_kv,
        ka: units.tm_ka,
        vintercept: units.volts,
    ):
        """
        :param motor_config: The specification data for your motor
        :param kv: Computed ``kv`` for your robot
        :param ka: Computed ``ka`` for your robot
        :param vintercept: The minimum voltage required to generate enough
                           torque to overcome steady-state friction (see the
                           paper for more details)
        """

        #: Current computed acceleration (in ft/s^2)
        self.acceleration = 0

        #: Current computed velocity (in ft/s)
        self.velocity = 0

        #: Current computed position (in ft)
        self.position = 0

        self._nominalVoltage = units.volts.m_from(
            motor_config.nominalVoltage,
            strict=False,
            name="motor_config.nominalVoltage",
        )
        self._vintercept = vintercept
        self._kv = kv
        self._ka = ka

    def compute(self, motor_pct: float, tm_diff: float) -> float:
        """
        :param motor_pct: Percentage of power for motor in range [1..-1]
        :param tm_diff:   Time elapsed since this function was last called

        :returns: velocity
        """

        appliedVoltage = self._nominalVoltage * motor_pct
        appliedVoltage = math.copysign(
            max(abs(appliedVoltage) - self._vintercept, 0), appliedVoltage
        )

        # Heun's method (taken from Ether's drivetrain calculator)
        # -> yn+1 = yn + (h/2) (f(xn, yn) + f(xn + h, yn +  h f(xn, yn)))
        a0 = self.acceleration
        v0 = self.velocity

        # initial estimate for next velocity/acceleration
        v1 = v0 + a0 * tm_diff
        a1 = (appliedVoltage - self._kv * v1) / self._ka

        # corrected trapezoidal estimate
        v1 = v0 + (a0 + a1) * 0.5 * tm_diff
        a1 = (appliedVoltage - self._kv * v1) / self._ka
        self.position += (v0 + v1) * 0.5 * tm_diff

        self.velocity = v1
        self.acceleration = a1

        return self.velocity


class TankModel:
    """
    This is a model of a FRC tankdrive-style drivetrain that will provide
    vaguely realistic motion for the simulator.

    This drivetrain model makes a number of assumptions:

    * N motors per side
    * Constant gearing
    * Motors are geared together
    * Wheels do not 'slip' on the ground
    * Each side of the robot moves in unison

    There are two ways to construct this model. You can use the theoretical
    model via :func:`TankModel.theory` and provide robot parameters
    such as gearing, total mass, etc.

    Alternatively, if you measure ``kv``, ``ka``, and ``vintercept`` as
    detailed in the paper mentioned above, you can plug those values in
    directly instead using the :class:`TankModel` constructor instead. For
    more information about measuring your own values, see the paper and
    `this thread on ChiefDelphi <https://www.chiefdelphi.com/forums/showthread.php?t=161539>`_.

    .. note:: You must specify the You can use whatever units you would like to specify the input
              parameter for your robot, RobotPy will convert them all
              to the correct units for computation.

              Output units for velocity and acceleration are in ft/s and
              ft/s^2

    Example usage for a 90lb robot with 2 CIM motors on each side with 6 inch
    wheels::

        from pyfrc.physics import motors, tankmodel
        from pyfrc.physics.units import units

        class PhysicsEngine:

            def __init__(self, physics_controller):
                self.physics_controller = physics_controller

                self.l_motor = hal.simulation.PWMSim(1)
                self.r_motor = hal.simulation.PWMSim(2)

                self.drivetrain = tankmodel.TankModel.theory(motors.MOTOR_CFG_CIM_IMP,
                                                             robot_mass=90 * units.lbs,
                                                             gearing=10.71, nmotors=2,
                                                             x_wheelbase=2.0*feet,
                                                             wheel_diameter=6*units.inch)

            def update_sim(self, now, tm_diff):
                l_motor = self.l_motor.getSpeed()
                r_motor = self.r_motor.getSpeed()

                transform = self.drivetrain.calculate(l_motor, r_motor, tm_diff)
                self.physics_controller.move_robot(transform)

                # optional: compute encoder
                # l_encoder = self.drivetrain.l_position * ENCODER_TICKS_PER_FT
                # r_encoder = self.drivetrain.r_position * ENCODER_TICKS_PER_FT
    """

    @classmethod
    def theory(
        cls,
        motor_config: MotorModelConfig,
        robot_mass: units.Quantity,
        gearing: float,
        nmotors: int = 1,
        x_wheelbase: units.Quantity = _kitbot_wheelbase,
        robot_width: units.Quantity = _kitbot_width,
        robot_length: units.Quantity = _kitbot_length,
        wheel_diameter: units.Quantity = 6 * units.inch,
        vintercept: units.volts = 1.3 * units.volts,
        timestep: int = 5 * units.ms,
    ):
        r"""
        Use this to create the drivetrain model when you haven't measured
        ``kv`` and ``ka`` for your robot.

        :param motor_config:    Specifications for your motor
        :param robot_mass:      Mass of the robot
        :param gearing:         Gear ratio .. so for a 10.74:1 ratio, you would pass 10.74
        :param nmotors:         Number of motors per side
        :param x_wheelbase:     Wheelbase of the robot
        :param robot_width:     Width of the robot
        :param robot_length:    Length of the robot
        :param wheel_diameter:  Diameter of the wheel
        :param vintercept:      The minimum voltage required to generate enough
                                torque to overcome steady-state friction (see the
                                paper for more details)
        :param timestep_ms:     Model computation timestep

        Computation of ``kv`` and ``ka`` are done as follows:

        * :math:`\omega_{free}` is the free speed of the motor
        * :math:`\tau_{stall}` is the stall torque of the motor
        * :math:`n` is the number of drive motors
        * :math:`m_{robot}` is the mass of the robot
        * :math:`d_{wheels}` is the diameter of the robot's wheels
        * :math:`r_{gearing}` is the total gear reduction between the motors and the wheels
        * :math:`V_{max}` is the nominal max voltage of the motor

        .. math::

            velocity_{max} = \frac{\omega_{free} \cdot \pi \cdot d_{wheels} }{r_{gearing}}

            acceleration_{max} = \frac{2 \cdot n \cdot \tau_{stall} \cdot r_{gearing} }{d_{wheels} \cdot m_{robot}}

            k_{v} = \frac{V_{max}}{velocity_{max}}

            k_{a} = \frac{V_{max}}{acceleration_{max}}
        """

        # Check input units
        # -> pint doesn't seem to support default args in check()
        Helpers.ensure_mass(robot_mass)
        Helpers.ensure_length(x_wheelbase)
        Helpers.ensure_length(robot_width)
        Helpers.ensure_length(robot_length)
        Helpers.ensure_length(wheel_diameter)

        max_velocity = (motor_config.freeSpeed * math.pi * wheel_diameter) / gearing
        max_acceleration = (2.0 * nmotors * motor_config.stallTorque * gearing) / (
            wheel_diameter * robot_mass
        )

        Helpers.ensure_velocity(max_velocity)
        Helpers.ensure_acceleration(max_acceleration)

        kv = motor_config.nominalVoltage / max_velocity
        ka = motor_config.nominalVoltage / max_acceleration

        kv = units.tm_kv.from_(kv, name="kv")
        ka = units.tm_ka.from_(ka, name="ka")

        logger.info(
            "Motor config: %d %s motors @ %.2f gearing with %.1f diameter wheels",
            nmotors,
            motor_config.name,
            gearing,
            wheel_diameter.m,
        )

        logger.info(
            "- Theoretical: vmax=%.3f ft/s, amax=%.3f ft/s^2, kv=%.3f, ka=%.3f",
            max_velocity.m_as(units.foot / units.second),
            max_acceleration.m_as(units.foot / units.second ** 2),
            kv.m,
            ka.m,
        )

        return cls(
            motor_config,
            robot_mass,
            x_wheelbase,
            robot_width,
            robot_length,
            kv,
            ka,
            vintercept,
            kv,
            ka,
            vintercept,
            timestep,
        )

    def __init__(
        self,
        motor_config: MotorModelConfig,
        robot_mass: units.Quantity,
        x_wheelbase: units.Quantity,
        robot_width: units.Quantity,
        robot_length: units.Quantity,
        l_kv: units.Quantity,
        l_ka: units.Quantity,
        l_vi: units.volts,
        r_kv: units.Quantity,
        r_ka: units.Quantity,
        r_vi: units.volts,
        timestep: units.Quantity = 5 * units.ms,
    ):
        """
        Use the constructor if you have measured ``kv``, ``ka``, and
        ``Vintercept`` for your robot. Use the :func:`.theory` function
        if you haven't.

        ``Vintercept`` is the minimum voltage required to generate enough
        torque to overcome steady-state friction (see the paper for more
        details).

        The robot width/length is used to compute the moment of inertia of
        the robot. Don't forget about bumpers!

        :param motor_config: Motor specification
        :param robot_mass:   Mass of robot
        :param x_wheelbase:  Wheelbase of the robot
        :param robot_width:  Width of the robot
        :param robot_length: Length of the robot
        :param l_kv:         Left side ``kv``
        :param l_ka:         Left side ``ka``
        :param l_vi:         Left side ``Vintercept``
        :param r_kv:         Right side ``kv``
        :param r_ka:         Right side ``ka``
        :param r_vi:         Right side ``Vintercept``
        :param timestep:     Model computation timestep
        """

        # check input parameters
        Helpers.ensure_mass(robot_mass)
        Helpers.ensure_length(x_wheelbase)
        Helpers.ensure_length(robot_width)
        Helpers.ensure_length(robot_length)
        Helpers.ensure_time(timestep)

        logger.info(
            "Robot base: %.1fx%.1f frame, %.1f wheelbase, %.1f mass",
            robot_width.m,
            robot_length.m,
            x_wheelbase.m,
            robot_mass.m,
        )

        self._lmotor = MotorModel(motor_config, l_kv, l_ka, l_vi)
        self._rmotor = MotorModel(motor_config, r_kv, r_ka, r_vi)

        self.inertia = (1 / 12.0) * robot_mass * (robot_length ** 2 + robot_width ** 2)

        # This is used to compute the rotational velocity
        self._bm = _bm_units.m_from((x_wheelbase / 2.0) * robot_mass)

        self._timestep = units.milliseconds.m_from(timestep, name="timestep") * 100

    @property
    def l_velocity(self):
        """The velocity of the left side (in ft/s)"""
        return self._lmotor.velocity

    @property
    def r_velocity(self):
        """The velocity of the right side (in ft/s)"""
        return self._rmotor.velocity

    @property
    def l_position(self):
        """The linear position of the left side wheel (in feet)"""
        return self._lmotor.position

    @property
    def r_position(self):
        """The linear position of the right side wheel (in feet)"""
        return self._rmotor.position

    @property
    def inertia(self):
        """
        The model computes a moment of inertia for your robot based on the
        given mass and robot width/length. If you wish to use a different
        moment of inertia, set this property after constructing the object

        Units are ``[mass] * [length] ** 2``
        """
        return self._inertia * _inertia_units

    @inertia.setter
    @units.wraps(None, (None, _inertia_units))
    def inertia(self, value):
        self._inertia = value

    def calculate(self, l_motor: float, r_motor: float, tm_diff: float) -> Transform2d:
        """
        Given motor values and the amount of time elapsed since this was last
        called, retrieves the x,y,angle that the robot has moved. Pass these
        values to :meth:`PhysicsInterface.distance_drive`.

        To update your encoders, use the ``l_position`` and ``r_position``
        attributes of this object.

        :param l_motor:    Left motor value (-1 to 1); 1 is forward
        :param r_motor:    Right motor value (-1 to 1); -1 is forward
        :param tm_diff:    Elapsed time since last call to this function

        :returns: transform containing x/y/angle offsets of robot travel

        .. note:: If you are using more than 2 motors, it is assumed that
                  all motors on each side are set to the same speed. Only
                  pass in one of the values from each side

        .. versionadded:: 2020.1.0
        """

        # This isn't quite right, the right way is to use matrix math. However,
        # this is Good Enough for now...
        x = 0
        y = 0
        angle = 0

        # split the time difference into timestep_ms steps
        total_time = int(tm_diff * 100000)
        steps = total_time // self._timestep
        remainder = total_time % self._timestep
        step = self._timestep / 100000.0
        if remainder:
            last_step = remainder / 100000.0
            steps += 1
        else:
            last_step = step

        while steps != 0:
            if steps == 1:
                tm_diff = last_step
            else:
                tm_diff = step

            steps -= 1

            l = self._lmotor.compute(l_motor, tm_diff)
            r = self._rmotor.compute(-r_motor, tm_diff)

            # Tank drive motion equations
            velocity = (l + r) * 0.5

            # Thanks to Tyler Veness for fixing the rotation equation, via conservation
            # of angular momentum equations
            # -> omega = b * m * (l - r) / J
            rotation = self._bm * (r - l) / self._inertia

            distance = velocity * tm_diff
            turn = rotation * tm_diff

            x += distance * math.cos(angle)
            y += distance * math.sin(angle)
            angle += turn

        return Transform2d.fromFeet(x, y, angle)



from math import radians as rad, pi

from pyfrc.physics.visionsim import VisionSimTarget


def test_visionsim_target1():

    # target is facing east, 90 degree viewing angle
    target = VisionSimTarget(0, 0, 315, 45)
    target.view_dst_start = 2
    target.view_dst_end = 4
    target.fov2 = rad(30)

    # check too close
    assert target.compute(0, 1, 1, 0) is None
    assert target.compute(0, -1, 1, 0) is None
    assert target.compute(0, 1, -1, 0) is None
    assert target.compute(0, -1, -1, 0) is None

    # check too far
    assert target.compute(0, 20, 20, 0) is None
    assert target.compute(0, -20, 20, 0) is None
    assert target.compute(0, 20, -20, 0) is None
    assert target.compute(0, -20, -20, 0) is None

    # check facing the right direction, outside of the view range
    assert target.compute(0, -2, -2, rad(-45)) is None
    assert target.compute(0, -2, 2, rad(45)) is None

    assert target.compute(0, 1, -2, rad(-135)) is None
    assert target.compute(0, 1, 2, rad(135)) is None

    # check facing the wrong direction, inside of the view range
    assert target.compute(0, 2, 2, rad(0)) is None
    assert target.compute(0, 2, -2, rad(0)) is None

    assert target.compute(0, 2, 2, rad(180)) is None
    assert target.compute(0, 2, -2, rad(180)) is None

    # check facing the right direction , inside of the view range
    result = target.compute(0, 2, 0, rad(180))
    assert result is not None
    assert result[0] == 1
    assert result[1] == 0
    assert result[2] == 0

    # TODO: add more tests here from different angles

    assert target.compute(0, 2, 2, rad(-135)) is not None
    assert target.compute(0, 2, -2, rad(135)) is not None

    assert target.compute(0, 2.07, -1.96, rad(140)) is not None
    # assert target.compute(0, 2, -2, rad(-135)) is not None


def test_visionsim_target2():

    # normalization function used by vision sim
    def _norm(angle):
        return ((rad(angle) + pi) % (pi * 2)) - pi

    # target is facing east, 90 degree viewing angle
    target = VisionSimTarget(18.5, 16, 315, 45)
    target.view_dst_start = 1.5
    target.view_dst_end = 15
    target.fov2 = rad(61.0 / 2.0)

    result = target.compute(0, 25, 16, _norm(-180))
    assert result is not None
    assert result[1] == 0

    result = target.compute(0, 25, 16, _norm(180))
    assert result is not None
    assert result[1] == 0

    # robot is northwest of target
    assert target.compute(0, 21.24, 13.82, _norm(206)) is None
    assert target.compute(0, 21.24, 13.82, _norm(149)) is not None
    assert target.compute(0, 21.24, 13.82, _norm(107)) is None
    assert target.compute(0, 21.24, 13.82, _norm(12.64)) is None
    assert target.compute(0, 21.24, 13.82, _norm(-136)) is None
    assert target.compute(0, 21.24, 13.82, _norm(-203)) is not None
    assert target.compute(0, 21.24, 13.82, _norm(943)) is None

    # robot is southwest of target
    assert target.compute(0, 20.22, 17.56, _norm(46)) is None
    assert target.compute(0, 20.22, 17.56, _norm(107)) is None
    assert target.compute(0, 20.22, 17.56, _norm(152.30)) is None
    assert target.compute(0, 20.22, 17.56, _norm(204.20)) is not None
    assert target.compute(0, 20.22, 17.56, _norm(-506.52)) is not None

    assert target.compute(0, 12.48, 13.79, _norm(24.61)) is None



import threading
import weakref


class TestRanTooLong(BaseException):
    """
    This is thrown when the time limit has expired

    This exception inherits from BaseException, so if you want to catch
    it you must explicitly specify it, as a blanket except statement
    will not catch this exception.'

    Generally, only internal pyfrc code needs to catch this
    """

    pass


class TestEnded(BaseException):
    """
    This is thrown when the controller has been signaled to end the test

    This exception inherits from BaseException, so if you want to catch
    it you must explicitly specify it, as a blanket except statement
    will not catch this exception.'

    Generally, only internal pyfrc code needs to catch this
    """

    pass


class TestFroze(BaseException):
    """
    This happens when an infinite loop of some kind in one of your
    non-robot threads is detected.
    """

    pass


class FakeTime:
    """
    Keeps track of time for robot code being tested, and makes sure the
    DriverStation is notified that new packets are coming in.

    .. note:: Do not create this object, your testing code can use this
              object to control time via the ``fake_time`` fixture
    """

    def __init__(self):
        # Note: when iterating this list, make copies or you'll run into errors
        self._child_threads = weakref.WeakKeyDictionary()
        self._children_free_run = False
        self._children_not_running = threading.Event()
        self._freeze_detect_threshold = 250
        self.lock = threading.RLock()

    def initialize(self):
        """
        Initializes fake time
        """

        self.reset()

        # Setup driver station hooks
        import wpilib

        assert not hasattr(wpilib.DriverStation, "instance"), (
            "You must not initialize the driver station before your robot "
            + "code executes. Perhaps you have a global somewhere? Globals are "
            + "generally evil and should be avoided!"
        )

        # The DS thread causes too many problems, disable it by getting
        # rid of the thread function
        wpilib.DriverStation._run = lambda _: None

        self._ds = wpilib.DriverStation.getInstance()

        # This is used to make DriverStation.waitForData() work
        self.ds_cond = _DSCondition(self)
        self._ds.waitForDataCond = self.ds_cond

        self.thread_id = threading.current_thread().ident
        return self.ds_cond

    def __time_test__(self):
        if self.time_limit is not None and self.time_limit <= self.time:
            raise TestRanTooLong()

    def _wake_children(self, timestep):
        # Subtract the timestep from the last requested times of the threads.
        # Wake them up if the requested time is in the past, and they haven't
        # already been stopped.
        waiting_on = []
        with self.lock:
            child_threads = list(self._child_threads.items())
        for thread, thread_info in child_threads:
            thread_info["time"] -= timestep
            if thread_info["time"] <= 0 and thread.is_alive():
                self._children_not_running.clear()
                thread_info["event"].set()  # Wake it up
                waiting_on.append(thread)

        # Wait for everything we woke up to sleep again.
        # Check that we did wake something up.
        if waiting_on:
            i = 0
            while not self._children_not_running.wait(0.020):
                i += 1
                if i == self._freeze_detect_threshold:
                    raise TestFroze("Waiting on %s" % waiting_on)

                # if this timed out, check to see if that particular
                # child died...
                for thread in waiting_on:
                    if thread.is_alive():
                        break  # out of the for loop, at least one child is active
                else:
                    break  # out of the while loop, everyone is dead

    def children_stopped(self):
        with self.lock:
            keys = list(self._child_threads.keys())
        for thread in keys:
            if thread.is_alive():
                return False
        return True

    def teardown(self):
        self._children_free_run = True  # Stop any threads being blocked
        self._children_not_running.set()  # Main thread needs to be running
        with self.lock:
            thread_infos = list(self._child_threads.values())
        for thread_info in thread_infos:
            thread_info["event"].set()

    def reset(self):
        """
        Resets the fake time to zero, and sets the time limit to default
        """
        self.time = 0
        self.time_limit = 500
        self.in_increment = False

        # Drop references to any child threads
        self._child_threads = weakref.WeakKeyDictionary()
        self._children_not_running.set()
        self._children_free_run = False

        self.next_ds_time = 0.020

    def get(self):
        """
        :returns: The current time for the robot
        """
        return self.time

    def increment_time_by(self, time):
        """
        Increments time by some number of seconds

        :param time: Number of seconds to increment time by
        :type time: float
        """
        # If it is a thread calling us, we intercept the call and insert
        # a blocking call to a threading.Event.wait() to force it to sleep
        # in the "real world", otherwise it keeps looping indeterminately.
        # Store the requested time, and at each DS step wake up those that
        # should have fired in the previous timestep.
        current_thread = threading.current_thread()
        if current_thread.ident != self.thread_id:
            with self.lock:
                child_info = self._child_threads.get(current_thread)
                if child_info is None:
                    child_info = {"event": threading.Event(), "time": time}
                    self._child_threads[current_thread] = child_info

            if time > 0 and not self._children_free_run:
                # Daughter thread needs to fire in the next DS time step,
                # so make it wait
                child_info["event"].clear()
                child_info["time"] = time
                # Check to see if the other threads have finished
                running = False
                with self.lock:
                    child_threads = list(self._child_threads.values())
                for v in child_threads:
                    if v["event"].is_set():
                        running = True
                if not running:
                    self._children_not_running.set()
                child_info["event"].wait()
            # We don't need to do anything else for children
            return

        if time < 0:
            time = 0

        with self.lock:
            if self.in_increment:
                raise ValueError(
                    "increment_time_by is not reentrant (did you call timer.delay?)"
                )

            next_ds = self.next_ds_time - self.time

        # When we wake up the TimerTasks, we can't do it while we have the lock
        # because some tasks need to get the time themselves
        # eg feeding motor watchdogs

        # This is terrible, but it works
        while time > 0:

            # Short wait, just get it done and exit
            if time < next_ds:
                with self.lock:
                    self.time += time
                    self.__time_test__()
                self._wake_children(time)
                break

            with self.lock:
                self.time += next_ds

                if current_thread.ident == self.thread_id:
                    self.ds_cond.on_step(self.time)

                # TODO: document this

                with self.ds_cond:
                    self.ds_cond.notify_all()
                    self._ds._getData()

                time -= next_ds

                next_ds = 0.020
                self.next_ds_time += next_ds

                self.__time_test__()
            self._wake_children(next_ds)

    def increment_new_packet(self):
        """
        Increment time enough to where the new DriverStation packet
        comes in
        """
        with self.lock:
            next_ds_time = self.next_ds_time
            time = self.time

        self.increment_time_by(next_ds_time - time)

    def set_time_limit(self, time_limit):
        """
        Sets the amount of time that a robot will run. When time is
        incremented past this time, a TestRanTooLong is thrown.

        The default time limit is 500 seconds

        :param time_limit: Number of seconds
        :type time_limit: float
        """
        self.time_limit = time_limit

    def _ds_time_offset(self):
        with self.lock:
            return self.next_ds_time - self.time


class _DSCondition(threading.Condition):
    """
    Condition variable replacement to allow fake time to be used to hook
    into the DriverStation packets.
    """

    def __init__(self, fake_time_inst, lock=None):
        super().__init__(lock)

        self.thread_id = threading.current_thread().ident
        self.fake_time_inst = fake_time_inst
        self._on_step = lambda tm: True

    def on_step(self, tm):

        retval = self._on_step(tm)
        if not retval:
            raise TestEnded()

    def wait(self, timeout=None):

        in_main = threading.current_thread().ident == self.thread_id

        # easy case: infinite timeout
        if timeout is None:

            # If on main thread, increment until the next packet time
            if in_main:
                self.fake_time_inst.increment_new_packet()
            else:
                # Not on main thread? wait for notify
                # -> TODO: this could never return if a notify doesn't occur
                super().wait()

            return True

        else:
            # harder case: some timeout
            # -> because this is the DS condition variable, we know when it will
            #    be notified. Determine if that time is before or after the next
            #    ds time
            ds_offset = self.fake_time_inst._ds_time_offset()

            # either way, increment by the time difference
            self.fake_time_inst.increment_time_by(ds_offset)

            # if before, return False indicating timeout
            # if after, return True indicating successful notification
            return timeout > ds_offset

        if timeout is not None:
            raise NotImplementedError("Didn't implement timeout support yet")

        # If we're on the robot's main thread, when this is called we just
        # need to increment the time, no wait required. If we're not on the
        # main thread, then we need to wait for a notification..

    # Copied from Python 3.6 source code, Python license
    # -> have to fork it because need to use our own time definition
    def wait_for(self, predicate, timeout=None):
        """Wait until a condition evaluates to True.

        predicate should be a callable which result will be interpreted as a
        boolean value.  A timeout may be provided giving the maximum time to
        wait.

        """
        endtime = None
        waittime = timeout
        result = predicate()
        while not result:
            if waittime is not None:
                if endtime is None:
                    endtime = self.fake_time_inst.get() + waittime
                else:
                    waittime = endtime - self.fake_time_inst.get()
                    if waittime <= 0:
                        break
            self.wait(waittime)
            result = predicate()
        return result



from collections import namedtuple

MotorModelConfig = namedtuple(
    "MotorModelConfig",
    [
        "name",
        "nominalVoltage",
        "freeSpeed",
        "freeCurrent",
        "stallTorque",
        "stallCurrent",
    ],
)
MotorModelConfig.__doc__ = """
    Configuration parameters useful for simulating a motor. Typically these
    parameters can be obtained from the manufacturer via a data sheet or other
    specification.
    
    RobotPy contains MotorModelConfig objects for many motors that are commonly
    used in FRC. If you find that we're missing a motor you care about, please
    file a bug report and let us know!
    
    .. note:: The motor configurations that come with pyfrc are defined using the
              pint units library. See :ref:`units`
    
"""
MotorModelConfig.name.__doc__ = "Descriptive name of motor"
MotorModelConfig.nominalVoltage.__doc__ = "Nominal voltage for the motor"
MotorModelConfig.freeSpeed.__doc__ = "No-load motor speed (``1 / [time]``)"
MotorModelConfig.freeCurrent.__doc__ = "No-load motor current"
MotorModelConfig.stallTorque.__doc__ = (
    "Stall torque (``[length]**2 * [mass] / [time]**2``)"
)
MotorModelConfig.stallCurrent.__doc__ = "Stall current"

from .units import units

NOMINAL_VOLTAGE = 12 * units.volts

#: Motor configuration for CIM
MOTOR_CFG_CIM = MotorModelConfig(
    "CIM",
    NOMINAL_VOLTAGE,
    5310 * units.cpm,
    2.7 * units.amps,
    2.42 * units.N_m,
    133 * units.amps,
)

#: Motor configuration for Mini CIM
MOTOR_CFG_MINI_CIM = MotorModelConfig(
    "MiniCIM",
    NOMINAL_VOLTAGE,
    5840 * units.cpm,
    3.0 * units.amps,
    1.41 * units.N_m,
    89.0 * units.amps,
)

#: Motor configuration for Bag Motor
MOTOR_CFG_BAG = MotorModelConfig(
    "Bag",
    NOMINAL_VOLTAGE,
    13180 * units.cpm,
    1.8 * units.amps,
    0.43 * units.N_m,
    53.0 * units.amps,
)

#: Motor configuration for 775 Pro
MOTOR_CFG_775PRO = MotorModelConfig(
    "775Pro",
    NOMINAL_VOLTAGE,
    18730 * units.cpm,
    0.7 * units.amps,
    0.71 * units.N_m,
    134 * units.amps,
)

#: Motor configuration for Andymark RS 775-125
MOTOR_CFG_775_125 = MotorModelConfig(
    "RS775-125",
    NOMINAL_VOLTAGE,
    5800 * units.cpm,
    1.6 * units.amps,
    0.28 * units.N_m,
    18.0 * units.amps,
)

#: Motor configuration for Banebots RS 775
MOTOR_CFG_BB_RS775 = MotorModelConfig(
    "RS775",
    NOMINAL_VOLTAGE,
    13050 * units.cpm,
    2.7 * units.amps,
    0.72 * units.N_m,
    97.0 * units.amps,
)

#: Motor configuration for Andymark 9015
MOTOR_CFG_AM_9015 = MotorModelConfig(
    "AM-9015",
    NOMINAL_VOLTAGE,
    14270 * units.cpm,
    3.7 * units.amps,
    0.36 * units.N_m,
    71.0 * units.amps,
)

#: Motor configuration for Banebots RS 550
MOTOR_CFG_BB_RS550 = MotorModelConfig(
    "RS550",
    NOMINAL_VOLTAGE,
    19000 * units.cpm,
    0.4 * units.amps,
    0.38 * units.N_m,
    84.0 * units.amps,
)

#: Motor configuration for NEO 550 Brushless Motor
MOTOR_CFG_NEO_550 = MotorModelConfig(
    "NEO 550",
    NOMINAL_VOLTAGE,
    11000 * units.cpm,
    1.4 * units.amps,
    0.97 * units.N_m,
    100 * units.amps,
)

#: Motor configuration for Falcon 500 Brushless Motor
MOTOR_CFG_FALCON_500 = MotorModelConfig(
    "Falcon 500",
    NOMINAL_VOLTAGE,
    6380 * units.cpm,
    1.5 * units.amps,
    4.69 * units.N_m,
    257 * units.amps,
)

del units



import contextlib

import hal
import wpilib

import pytest

from pyfrc.test_support.pytest_plugin import PyFrcPlugin
from pyfrc.tests.basic import test_practice as _test_practice


@contextlib.contextmanager
def get_plugin(cls):

    wpilib.DriverStation._reset()

    plugin = PyFrcPlugin(cls, None, None)
    plugin.pytest_runtest_setup()

    try:
        yield plugin, plugin.get_control()
    finally:
        plugin.pytest_runtest_teardown(None)


class SampleRobotInit(wpilib.SampleRobot):
    def robotInit(self):
        assert False


class SampleAutonomous(wpilib.SampleRobot):
    def autonomous(self):
        assert False


class SampleTeleop(wpilib.SampleRobot):
    def operatorControl(self):
        assert False


class IterativeRobotInit(wpilib.IterativeRobot):
    def robotInit(self):
        assert False


class IterativeAutoInit(wpilib.IterativeRobot):
    def autonomousInit(self):
        assert False


class IterativeAutoPeriodic(wpilib.IterativeRobot):
    def autonomousPeriodic(self):
        assert False


class IterativeTeleopInit(wpilib.IterativeRobot):
    def teleopInit(self):
        assert False


class IterativeTeleopPeriodic(wpilib.IterativeRobot):
    def teleopPeriodic(self):
        assert False


@pytest.mark.parametrize(
    "cls",
    [
        SampleRobotInit,
        SampleAutonomous,
        IterativeRobotInit,
        IterativeAutoInit,
        IterativeAutoPeriodic,
    ],
)
def test_auto_failure(cls):
    """Ensure that failures can be detected in autonomous mode"""
    with get_plugin(cls) as (plugin, control):
        control.set_autonomous(enabled=True)
        with pytest.raises(AssertionError):
            control.run_test(lambda tm: tm < 15)


@pytest.mark.parametrize(
    "cls",
    [
        SampleRobotInit,
        SampleTeleop,
        IterativeRobotInit,
        IterativeTeleopInit,
        IterativeTeleopPeriodic,
    ],
)
def test_teleop_failure(cls):
    """Ensure that failures can be detected in teleop mode"""
    with get_plugin(cls) as (plugin, control):
        control.set_operator_control(enabled=True)
        with pytest.raises(AssertionError):
            control.run_test(lambda tm: tm < 15)


class Iterative(wpilib.IterativeRobot):
    def robotInit(self):
        self.did_robot_init = True

    def disabledInit(self):
        self.did_disabled_init = True

    def disabledPeriodic(self):
        self.did_disabled_periodic = True

    def autonomousInit(self):
        self.did_auto_init = True

    def autonomousPeriodic(self):
        self.did_auto_periodic = True

    def teleopInit(self):
        self.did_teleop_init = True

    def teleopPeriodic(self):
        self.did_teleop_periodic = True


def test_iterative():
    """Ensure that all states of the iterative robot run"""
    with get_plugin(Iterative) as (plugin, control):
        _test_practice(plugin.get_control(), plugin.get_fake_time(), plugin.get_robot())

        robot = plugin.get_robot()
        assert robot.did_robot_init == True
        assert robot.did_disabled_init == True
        assert robot.did_disabled_periodic == True
        assert robot.did_auto_init == True
        assert robot.did_auto_periodic == True
        assert robot.did_teleop_init == True
        assert robot.did_teleop_periodic == True


class Sample(wpilib.SampleRobot):
    def robotInit(self):
        self.did_robot_init = True

    def disabled(self):
        self.did_robot_disabled = True

    def autonomous(self):
        self.did_autonomous = True

    def operatorControl(self):
        self.did_operator = True


def test_sample():
    """Ensure that all states of the sample robot run"""
    with get_plugin(Sample) as (plugin, control):
        _test_practice(plugin.get_control(), plugin.get_fake_time(), plugin.get_robot())

        robot = plugin.get_robot()
        assert robot.did_robot_init == True
        assert robot.did_robot_disabled == True
        assert robot.did_autonomous == True
        assert robot.did_operator == True



import argparse
import contextlib
import inspect
import os
import sys
import re

import shutil
import tempfile
import threading

from os.path import abspath, basename, dirname, join, splitext
from pathlib import PurePosixPath

from robotpy_installer import sshcontroller

from ..util import print_err, yesno

import wpilib

import logging

logger = logging.getLogger("deploy")


def relpath(path):
    """Path helper, gives you a path relative to this file"""
    return os.path.normpath(
        os.path.join(os.path.abspath(os.path.dirname(__file__)), path)
    )


@contextlib.contextmanager
def wrap_ssh_error(msg: str):
    try:
        yield
    except sshcontroller.SshExecError as e:
        raise sshcontroller.SshExecError(f"{msg}: {str(e)}", e.retval) from e


class PyFrcDeploy:
    """
    Uploads your robot code to the robot and executes it immediately
    """

    def __init__(self, parser: argparse.ArgumentParser):
        parser.add_argument(
            "--builtin",
            default=False,
            action="store_true",
            help="Use pyfrc's builtin tests if no tests are specified",
        )

        parser.add_argument(
            "--skip-tests",
            action="store_true",
            default=False,
            help="If specified, don't run tests before uploading code to robot (DANGEROUS)",
        )

        parser.add_argument(
            "--debug",
            action="store_true",
            default=False,
            help="If specified, runs the code in debug mode (which only currently enables verbose logging)",
        )

        parser.add_argument(
            "--nonstandard",
            action="store_true",
            default=False,
            help="When specified, allows you to deploy code in a file that isn't called robot.py",
        )

        parser.add_argument(
            "--nc",
            "--netconsole",
            action="store_true",
            default=False,
            help="Attach netconsole listener and show robot stdout (requires DS to be connected)",
        )

        parser.add_argument(
            "--nc-ds",
            "--netconsole-ds",
            action="store_true",
            default=False,
            help="Attach netconsole listener and show robot stdout (fakes a DS connection)",
        )

        parser.add_argument(
            "--in-place",
            action="store_true",
            default=False,
            help="Overwrite currently deployed code, don't delete anything, and don't restart running robot code.",
        )

        parser.add_argument(
            "-n",
            "--no-version-check",
            action="store_true",
            default=False,
            help="If specified, don't verify that your local wpilib install matches the version on the robot (not recommended)",
        )

        parser.add_argument(
            "--large",
            action="store_true",
            default=False,
            help="If specified, allow uploading large files (> 250k) to the RoboRIO",
        )

        robot_args = parser.add_mutually_exclusive_group()

        robot_args.add_argument(
            "--robot", default=None, help="Set hostname or IP address of robot"
        )

        robot_args.add_argument(
            "--team", default=None, type=int, help="Set team number to deploy robot for"
        )

        parser.add_argument(
            "--no-resolve",
            action="store_true",
            default=False,
            help="If specified, don't do a DNS lookup, allow ssh et al to do it instead",
        )

    def run(self, options, robot_class, **static_options):

        from .. import config

        config.mode = "upload"

        # run the test suite before uploading
        # TODO: disabled for 2020
        if False and not options.skip_tests:
            from .cli_test import PyFrcTest

            tester = PyFrcTest()

            retval = tester.run_test(
                [], robot_class, options.builtin, ignore_missing_test=True
            )
            if retval != 0:
                print_err("ERROR: Your robot tests failed, aborting upload.")
                if not sys.stdin.isatty():
                    print_err("- Use --skip-tests if you want to upload anyways")
                    return retval

                print()
                if not yesno("- Upload anyways?"):
                    return retval

                if not yesno("- Are you sure? Your robot code may crash!"):
                    return retval

                print()
                print("WARNING: Uploading code against my better judgement...")

        # upload all files in the robot.py source directory
        robot_file = abspath(inspect.getfile(robot_class))
        robot_path = dirname(robot_file)
        robot_filename = basename(robot_file)
        cfg_filename = join(robot_path, ".deploy_cfg")

        if not options.nonstandard and robot_filename != "robot.py":
            print_err(
                "ERROR: Your robot code must be in a file called robot.py (launched from %s)!"
                % robot_filename
            )
            print_err()
            print_err(
                "If you really want to do this, then specify the --nonstandard argument"
            )
            return 1

        if not options.large and not self._check_large_files(robot_path):
            return 1

        hostname_or_team = options.robot
        if not hostname_or_team and options.team:
            hostname_or_team = options.team

        try:
            with sshcontroller.ssh_from_cfg(
                cfg_filename,
                username="lvuser",
                password="",
                hostname=hostname_or_team,
                no_resolve=options.no_resolve,
            ) as ssh:

                if not self._check_requirements(ssh, options.no_version_check):
                    return 1

                if not self._do_deploy(ssh, options, robot_filename, robot_path):
                    return 1

        except sshcontroller.SshExecError as e:
            print_err("ERROR:", str(e))
            return 1

        print("\nSUCCESS: Deploy was successful!")
        return 0

    def _check_large_files(self, robot_path):

        large_sz = 250000

        large_files = []
        for fname in self._copy_to_tmpdir(None, robot_path, dry_run=True):
            st = os.stat(fname)
            if st.st_size > large_sz:
                large_files.append((fname, st.st_size))

        if large_files:
            print_err(f"ERROR: large files found (larger than {large_sz} bytes)")
            for fname, sz in sorted(large_files):
                print_err(f"- {fname} ({sz} bytes)")

            if not yesno("Upload anyways?"):
                return False

        return True

    def _check_requirements(
        self, ssh: sshcontroller.SshController, no_wpilib_version_check: bool
    ) -> bool:

        # does python exist
        with wrap_ssh_error("checking if python exists"):
            if ssh.exec_cmd("[ -x /usr/local/bin/python3 ]").returncode != 0:
                print_err(
                    "ERROR: python3 was not found on the roboRIO: have you installed robotpy?"
                )
                print_err()
                print_err(
                    f"See {sys.executable} -m robotpy-installer install-python --help"
                )
                return False

        # does wpilib exist and does the version match
        with wrap_ssh_error("checking for wpilib version"):
            py = ";".join(
                [
                    "import os.path, site",
                    "version = 'unknown'",
                    "v = site.getsitepackages()[0] + '/wpilib/version.py'",
                    "exec(open(v).read(), globals()) if os.path.exists(v) else False",
                    "print(version)",
                ]
            )

            result = ssh.exec_cmd(
                f'/usr/local/bin/python3 -c "{py}"', check=True, get_output=True
            )
            assert result.stdout is not None

            wpilib_version = result.stdout.strip()
            if wpilib_version == "unknown":
                print_err(
                    "WPILib was not found on the roboRIO: have you installed it on the RoboRIO?"
                )
                return False

            print("RoboRIO has WPILib version", wpilib_version)

            if not no_wpilib_version_check and wpilib_version != wpilib.__version__:
                print_err("ERROR: expected WPILib version %s" % wpilib.__version__)
                print_err()
                print_err("You should either:")
                print_err(
                    "- If the robot version is older, upgrade the RobotPy on your robot"
                )
                print_err("- Otherwise, upgrade pyfrc on your computer")
                print_err()
                print_err(
                    "Alternatively, you can specify --no-version-check to skip this check"
                )
                return False

        # does startup dlls need to be fixed
        with wrap_ssh_error("checking StartupDLLs"):
            result = ssh.exec_cmd(
                "grep ^StartupDLLs /etc/natinst/share/ni-rt.ini",
                get_output=True,
            )
            assert result.stdout is not None
            if result.stdout.strip() != "":
                self._fix_startupdlls(ssh)

        return True

    def _fix_startupdlls(self, lvuser_ssh: sshcontroller.SshController):
        # requires admin-level access to fix, so create a new ssh controller
        ssh = sshcontroller.SshController(lvuser_ssh.hostname, "admin", "")
        with ssh:
            ssh.exec_cmd(
                'sed -i -e "s/^StartupDLLs/;StartupDLLs/" /etc/natinst/share/ni-rt.ini'
            )

    def _do_deploy(
        self,
        ssh: sshcontroller.SshController,
        options,
        robot_filename: str,
        robot_path: str,
    ) -> bool:
        # This probably should be configurable... oh well

        deploy_dir = PurePosixPath("/home/lvuser")
        py_deploy_subdir = "py"
        py_new_deploy_subdir = "py_new"
        py_deploy_dir = deploy_dir / py_deploy_subdir

        # note below: deployed_cmd appears that it only can be a single line

        # In 2015, there were stdout/stderr issues. In 2016+, they seem to
        # have been fixed, but need to use -u for it to really work properly

        if options.debug:
            compileall_flags = ""
            deployed_cmd = (
                "env LD_LIBRARY_PATH=/usr/local/frc/lib/ /usr/local/bin/python3 -u %s/%s -v run"
                % (py_deploy_dir, robot_filename)
            )
            deployed_cmd_fname = "robotDebugCommand"
            bash_cmd = "/bin/bash -cex"
        else:
            compileall_flags = "-O"
            deployed_cmd = (
                "env LD_LIBRARY_PATH=/usr/local/frc/lib/ /usr/local/bin/python3 -u -O %s/%s run"
                % (py_deploy_dir, robot_filename)
            )
            deployed_cmd_fname = "robotCommand"
            bash_cmd = "/bin/bash -ce"

        if options.in_place:
            replace_cmd = "true"
            py_new_deploy_subdir = py_deploy_subdir
        else:
            replace_cmd = (
                "rm -rf %(py_deploy_dir)s; mv %(py_new_deploy_dir)s %(py_deploy_dir)s"
            )

        py_new_deploy_dir = deploy_dir / py_new_deploy_subdir
        replace_cmd %= {
            "py_deploy_dir": py_deploy_dir,
            "py_new_deploy_dir": py_new_deploy_dir,
        }

        with wrap_ssh_error("configuring command"):
            ssh.exec_cmd(
                f'echo "{deployed_cmd}" > {deploy_dir}/{deployed_cmd_fname}', check=True
            )

        if options.debug:
            with wrap_ssh_error("touching frcDebug"):
                ssh.exec_cmd("touch /tmp/frcdebug", check=True)

        with wrap_ssh_error("removing stale deploy directory"):
            ssh.exec_cmd(f"rm -rf {py_new_deploy_dir}", check=True)

        # Copy the files over, copy to a temporary directory first
        # -> this is inefficient, but it's easier in sftp
        tmp_dir = tempfile.mkdtemp()
        try:
            py_tmp_dir = join(tmp_dir, py_new_deploy_subdir)
            self._copy_to_tmpdir(py_tmp_dir, robot_path)
            ssh.sftp(py_tmp_dir, deploy_dir, mkdir=not options.in_place)
        finally:
            shutil.rmtree(tmp_dir)

        # start the netconsole listener now if requested, *before* we
        # actually start the robot code, so we can see all messages
        nc_thread = None
        if options.nc or options.nc_ds:
            nc_thread = self._start_nc(ssh, options)

        if not options.in_place:
            # Restart the robot code and we're done!
            sshcmd = (
                "%(bash_cmd)s '"
                + "%(replace_cmd)s;"
                + "/usr/local/bin/python3 %(compileall_flags)s -m compileall -q -r 5 /home/lvuser/py;"
                + ". /etc/profile.d/natinst-path.sh; "
                + "chown -R lvuser:ni %(py_deploy_dir)s; "
                + "sync; "
                + "/usr/local/frc/bin/frcKillRobot.sh -t -r || true"
                + "'"
            )

            sshcmd %= {
                "bash_cmd": bash_cmd,
                "compileall_flags": compileall_flags,
                "py_deploy_dir": py_deploy_dir,
                "replace_cmd": replace_cmd,
            }

            logger.debug("SSH: %s", sshcmd)

            with wrap_ssh_error("starting robot code"):
                ssh.exec_cmd(sshcmd, check=True, print_output=True)

        if nc_thread is not None:
            nc_thread.join()

        return True

    def _start_nc(self, ssh, options):
        from netconsole import run

        nc_event = threading.Event()
        nc_thread = threading.Thread(
            target=run,
            args=(ssh.hostname,),
            kwargs=dict(connect_event=nc_event, fakeds=options.nc_ds),
            daemon=True,
        )
        nc_thread.start()
        nc_event.wait(5)
        logger.info("Netconsole is listening...")
        return nc_thread

    def _copy_to_tmpdir(self, tmp_dir, robot_path, dry_run=False):

        upload_files = []
        ignore_exts = {"pyc", "whl", "ipk", "zip", "gz"}

        for root, dirs, files in os.walk(robot_path):
            prefix = root[len(robot_path) + 1 :]
            if not dry_run:
                os.mkdir(join(tmp_dir, prefix))

            # skip .svn, .git, .hg, etc directories
            for d in dirs[:]:
                if d.startswith(".") or d in ("__pycache__", "venv"):
                    dirs.remove(d)

            # skip .pyc files
            for filename in files:

                r, ext = splitext(filename)
                if ext in ignore_exts or r.startswith("."):
                    continue

                fname = join(root, filename)
                upload_files.append(fname)

                if not dry_run:
                    shutil.copy(fname, join(tmp_dir, prefix, filename))

        return upload_files



import inspect
import os
import re
import sys

# if you want to be really pedantic, enforce sphinx docstrings. Ha.
pedantic_docstrings = True

# regex to use to detect the sphinx docstrings
param_re = re.compile(r"^:param (\S+?):\s*(.+)$")
type_re = re.compile(r"^:type (\S+?):\s*(.+)$")
both_re = re.compile(r"^:param (\S+?) (\S+?):\s*(.+)$")


def ignore_object(o, robot_path):
    """Returns true if the object can be ignored"""

    if inspect.isbuiltin(o):
        return True

    try:
        src = inspect.getsourcefile(o)
    except TypeError:
        return True

    return src is None or not os.path.abspath(src).startswith(robot_path)


def print_fn_err(msg, parent, fn, errors):
    if inspect.isclass(parent):
        name = "%s.%s" % (parent.__name__, fn.__name__)
    else:
        name = "%s" % fn.__name__
    name = "%s.%s" % (parent.__name__, fn.__name__)

    err = "ERROR: %s '%s()'\n-> See %s:%s" % (
        msg,
        name,
        inspect.getsourcefile(fn),
        inspect.getsourcelines(fn)[1],
    )

    if err not in errors:
        print(err, file=sys.stderr)
        errors.append(err)


def check_function(parent, fn, errors):
    doc = inspect.getdoc(fn)

    if doc is None:
        print_fn_err("No docstring for", parent, fn, errors)

    elif pedantic_docstrings:
        # find the list of parameters
        args, varargs, keywords, defaults, _, _, annotations = inspect.getfullargspec(
            fn
        )
        if len(args) > 0 and args[0] == "self":
            del args[0]

        if varargs is not None:
            args.append(varargs)

        if keywords is not None:
            args.append(keywords)

        params = []

        for line in doc.splitlines():
            # :param arg: stuff
            match = param_re.match(line)
            if match:

                arg = match.group(1)
                if arg not in args:
                    print_fn_err(
                        "Param '%s' is documented but isn't a parameter for" % arg,
                        parent,
                        fn,
                        errors,
                    )

                if arg in annotations.keys() and isinstance(annotations.get(arg), str):
                    print_fn_err(
                        "Do not document %s in both the docstring and annotations"
                        % arg,
                        parent,
                        fn,
                        errors,
                    )

                params.append(arg)

            # :type arg: type
            match = type_re.match(line)
            if match:
                arg = match.group(1)
                if arg in annotations.keys() and isinstance(annotations.get(arg), type):
                    print_fn_err(
                        "Do not document %s in both the docstring and annotations"
                        % arg,
                        parent,
                        fn,
                        errors,
                    )

            # :param type arg: stuff
            match = both_re.match(line)
            if match:
                arg = match.group(2)
                if arg not in args:
                    print_fn_err(
                        "Param '%s' is documented but isn't a parameter for" % arg,
                        parent,
                        fn,
                        errors,
                    )

                if arg in annotations.keys() and isinstance(
                    annotations.get(arg), (str, type)
                ):
                    print_fn_err(
                        "Do not document %s in both the docstring and annotations"
                        % arg,
                        parent,
                        fn,
                        errors,
                    )

                params.append(arg)

        for param, annotation in annotations.items():
            if param not in params and isinstance(annotation, str):
                params.insert(args.index(param), param)

        if len(params) != len(args):

            diff = set(args).difference(params)

            if len(diff) == 1:
                print_fn_err(
                    "Param '%s' is not documented in docstring for" % diff.pop(),
                    parent,
                    fn,
                    errors,
                )
            elif len(diff) > 1:
                print_fn_err(
                    "Params '%s' are not documented in docstring for"
                    % "','".join(diff),
                    parent,
                    fn,
                    errors,
                )

        else:
            for param, arg in zip(params, args):
                if param != arg:
                    print_fn_err(
                        "Param '%s' is out of order, does not match param '%s' in docstring for"
                        % (param, arg),
                        parent,
                        fn,
                        errors,
                    )


def check_object(o, robot_path, errors):
    if inspect.isclass(o) and inspect.getdoc(o) is None:
        err = "ERROR: Class '%s' has no docstring!\n-> See %s:%s" % (
            o.__name__,
            inspect.getsourcefile(o),
            inspect.getsourcelines(o)[1],
        )
        print(err)
        errors.append(err)

    for name, value in inspect.getmembers(o):

        if ignore_object(value, robot_path):
            continue

        check_thing(o, value, robot_path, errors)


def check_thing(parent, thing, robot_path, errors):
    if inspect.isclass(thing):
        check_object(thing, robot_path, errors)

    elif inspect.isfunction(thing):
        check_function(parent, thing, errors)


def test_docstrings(robot, robot_path):
    """
    The purpose of this test is to ensure that all of your robot code
    has docstrings. Properly using docstrings will make your code
    more maintainable and look more professional.
    """

    # this allows abspath() to work correctly
    os.chdir(robot_path)

    errors = []

    for module in sys.modules.values():

        if ignore_object(module, robot_path):
            continue

        check_object(module, robot_path, errors)

    # if you get an error here, look at stdout for the error message
    assert len(errors) == 0



"""
    The 'vision simulator' provides objects that assist in modeling inputs
    from a camera processing system.
"""

import collections
import math

inf = float("inf")
twopi = math.pi * 2.0


def _in_angle_interval(x, a, b):
    return (x - a) % twopi <= (b - a) % twopi


class VisionSimTarget:
    """
    Target object that you pass the to the constructor of :class:`.VisionSim`
    """

    def __init__(self, x, y, view_angle_start, view_angle_end):
        """
        :param x: Target x position
        :param y: Target y position

        :param view_angle_start:
        :param view_angle_end: clockwise from start angle

        View angle is defined in degrees from 0 to 360, with 0 = east,
        increasing clockwise. So, if the robot could only see
        the target from the south east side, you would use a view angle
        of start=0, end=90.
        """
        self.x = x
        self.y = y

        self.view_angle_start = math.radians(view_angle_start)
        self.view_angle_end = math.radians(view_angle_end)

    def compute(self, now, x, y, angle):

        # incoming angle must be normalized to -180,180 (done in VisionSim)

        # note: the position tracking of the robot has Y incrementing
        # positively downwards, presumably because it's convenient for
        # drawing stuff. Not what I had expected.

        # compute target distance
        dx = self.x - x
        dy = self.y - y

        distance = math.hypot(dx, dy)

        # at the right distance?
        if distance < self.view_dst_start or distance > self.view_dst_end:
            return

        # determine the absolute angle of the target relative to the robot
        target_angle = math.atan2(dy, dx)

        # in the view range?
        if not _in_angle_interval(
            target_angle + math.pi, self.view_angle_start, self.view_angle_end
        ):
            return

        # is the robot facing the target?
        a = angle - self.fov2
        b = angle + self.fov2

        if _in_angle_interval(target_angle + math.pi, a + math.pi, b + math.pi):
            offset = math.degrees(
                (((target_angle - angle) + math.pi) % (math.pi * 2)) - math.pi
            )
            return (1, now, offset, distance)


class VisionSim:
    """
    This helper object is designed to help you simulate input from a
    vision system. The algorithm is a very simple approximation and
    has some weaknesses, but it should be a good start and general
    enough to work for many different usages.

    There are a few assumptions that this makes:

    - Your camera code sends new data at a constant frequency
    - The data from the camera lags behind at a fixed latency
    - If the camera is too close, the target cannot be seen
    - If the camera is too far, the target cannot be seen
    - You can only 'see' the target when the 'front' of the robot is
      around particular angles to the target
    - The camera is in the center of your robot (this simplifies some
      things, maybe fix this in the future...)

    To use this, create an instance in your physics simulator::

        targets = [
            VisionSim.Target(...)
        ]


    Then call the :meth:`compute` method from your ``update_sim`` method
    whenever your camera processing is enabled::

        # in physics engine update_sim()
        x, y, angle = self.physics_controller.get_position()

        if self.camera_enabled:
            data = self.vision_sim.compute(now, x, y, angle)
            if data is not None:
                self.nt.putNumberArray('/camera/target', data[0])
        else:
            self.vision_sim.dont_compute()

    .. note:: There is a working example in the examples repository
              you can use to try this functionality out
    """

    Target = VisionSimTarget

    def __init__(
        self,
        targets,
        camera_fov,
        view_dst_start,
        view_dst_end,
        data_frequency=15,
        data_lag=0.050,
        physics_controller=None,
    ):
        """
        There are a lot of constructor parameters:

        :param targets:          List of target positions (x, y) on field in feet
        :param view_angle_start: Center angle that the robot can 'see' the target from (in degrees)
        :param camera_fov:       Field of view of camera (in degrees)
        :param view_dst_start:   If the robot is closer than this, the target cannot be seen
        :param view_dst_end:     If the robot is farther than this, the target cannot be seen
        :param data_frequency:   How often the camera transmits new coordinates
        :param data_lag:         How long it takes for the camera data to be processed
                                 and make it to the robot
        :param physics_controller: If set, will draw target information in UI
        """

        fov2 = math.radians(camera_fov / 2.0)

        self.distance = 0
        self.update_period = 1.0 / data_frequency
        self.data_lag = data_lag

        self.last_compute_time = -10
        self.send_queue = collections.deque()

        self.targets = targets

        assert view_dst_start < view_dst_end
        assert self.data_lag > 0.001

        # objects = []
        # if physics_controller:
        #     objects = physics_controller.config_obj["pyfrc"]["field"]["objects"]

        for target in targets:
            target.camera_fov = camera_fov
            target.view_dst_start = view_dst_start
            target.view_dst_end = view_dst_end
            target.fov2 = fov2

            # objects.append(
            #     {"color": "red", "rect": [target.x - 0.1, target.y - 0.1, 0.4, 0.4]}
            # )

    def dont_compute(self):
        """
        Call this when vision processing should be disabled
        """
        self.send_queue.clear()

    def get_immediate_distance(self):
        """
        Use this data to feed to a sensor that is mostly instantaneous
        (such as an ultrasonic sensor).

        .. note:: You must call :meth:`compute` first.
        """
        return self.distance

    def compute(self, now, x, y, angle):
        """
        Call this when vision processing should be enabled

        :param now:   The value passed to ``update_sim``
        :param x:     Returned from physics_controller.get_position
        :param y:     Returned from physics_controller.get_position
        :param angle: Returned from physics_controller.get_position

        :returns: None or list of tuples of (found=0 or 1, capture_time, offset_degrees, distance).
                  The tuples are ordered by absolute offset from the
                  target. If a list is returned, it is guaranteed to have at
                  least one element in it.

                  Note: If your vision targeting doesn't have the ability
                  to focus on multiple targets, then you should ignore
                  the other elements.
        """

        # Normalize angle to [-180,180]
        output = []
        angle = ((angle + math.pi) % (math.pi * 2)) - math.pi

        for target in self.targets:
            proposed = target.compute(now, x, y, angle)
            if proposed:
                output.append(proposed)

        if not output:
            output.append((0, now, inf, 0))
            self.distance = None
        else:
            # order by absolute offset
            output.sort(key=lambda i: abs(i[2]))
            self.distance = output[-1][3]

        # Only store stuff every once in awhile
        if now - self.last_compute_time > self.update_period:

            self.last_compute_time = now
            self.send_queue.appendleft(output)

        # simulate latency by delaying camera output
        if self.send_queue:
            output = self.send_queue[-1]
            if now - output[0][1] > self.data_lag:
                return self.send_queue.pop()



import inspect
import json
from os import mkdir
from os.path import abspath, dirname, exists, join

physics_starter = '''
#
# See the documentation for more details on how this works
#
# Documentation can be found at https://robotpy.readthedocs.io/projects/pyfrc/en/latest/physics.html
#
# The idea here is you provide a simulation object that overrides specific
# pieces of WPILib, and modifies motors/sensors accordingly depending on the
# state of the simulation. An example of this would be measuring a motor
# moving for a set period of time, and then changing a limit switch to turn
# on after that period of time. This can help you do more complex simulations
# of your robot code without too much extra effort.
#
# Examples can be found at https://github.com/robotpy/examples


class PhysicsEngine:
    """
       Simulates a 4-wheel robot using Tank Drive joystick control
    """

    def __init__(self, physics_controller):
        """
        :param physics_controller: `pyfrc.physics.core.Physics` object
                                   to communicate simulation effects to
        """

        self.physics_controller = physics_controller

        """
        # Change these parameters to fit your robot!
        bumper_width = 3.25 * units.inch

        self.drivetrain = tankmodel.TankModel.theory(
            motor_cfgs.MOTOR_CFG_CIM,           # motor configuration
            110 * units.lbs,                    # robot mass
            10.71,                              # drivetrain gear ratio \omega In/\omega Out
            2,                                  # motors per side
            22 * units.inch,                    # robot wheelbase
            23 * units.inch + bumper_width * 2, # robot width
            32 * units.inch + bumper_width * 2, # robot length
            6 * units.inch,                     # wheel diameter
        )
        """

    def update_sim(self, hal_data, now, tm_diff):
        """
        Called when the simulation parameters for the program need to be
        updated.

        :param now: The current time as a float
        :param tm_diff: The amount of time that has passed since the last
                        time that this function was called
        """

        # Simulate the drivetrain
        """
        lr_motor = hal_data["pwm"][1]["value"]
        rr_motor = hal_data["pwm"][2]["value"]

        # Not needed because front and rear should be in sync
        # lf_motor = hal_data['pwm'][3]['value']
        # rf_motor = hal_data['pwm'][4]['value']

        x, y, angle = self.drivetrain.get_distance(lr_motor, rr_motor, tm_diff)
        self.physics_controller.distance_drive(x, y, angle)
        """
'''

sim_starter = {
    "pyfrc": {
        "robot": {
            "w": 3,
            "h": 2,
            "starting_x": 10,
            "starting_y": 10,
            "starting_angle": 0,
        },
        "pwm": {},
        "can": {},
        "joysticks": {
            "0": {"axes": {}, "buttons": {}},
            "1": {"axes": {}, "buttons": {}},
        },
    }
}


class PyFrcCreatePhysics:
    def __init__(self, parser=None):
        pass

    def run(self, options, robot_class, **static_options):

        print("Robot physics are not yet implemented for RobotPy 2020")
        return 1

        robot_file = abspath(inspect.getfile(robot_class))
        robot_path = dirname(robot_file)
        sim_path = join(robot_path, "sim")

        physics_file = join(robot_path, "physics.py")
        sim_file = join(sim_path, "config.json")
        if exists(physics_file):
            print("- physics.py already exists")
        else:
            with open(physics_file, "w") as fp:
                fp.write(physics_starter)
            print("- physics file created at", physics_file)

        if exists(sim_path):
            if exists(sim_file):
                print("- config.json already exists")
            else:
                with open(sim_file, "w") as f:
                    f.write(json.dumps(sim_starter, indent=4) + "\n")
                print("- sim file created at", sim_file)
        else:
            mkdir(sim_path)
            with open(sim_file, "w") as f:
                f.write(json.dumps(sim_starter, indent=4) + "\n")
            print("- sim file created at", sim_file)
        print()
        print("Robot simulation can be run via 'python3 robot.py sim'")



import inspect
import os
from os.path import abspath, dirname, exists, join

builtin_tests = """'''
    This test module imports tests that come with pyfrc, and can be used
    to test basic functionality of just about any robot.
'''

from pyfrc.tests import *
"""


class PyFrcAddTests:
    def __init__(self, parser=None):
        pass

    def run(self, options, robot_class, **static_options):

        print("tests are not yet implemented for RobotPy 2020")
        return 1

        robot_file = abspath(inspect.getfile(robot_class))
        robot_path = dirname(robot_file)

        try_dirs = [
            abspath(join(robot_path, "tests")),
            abspath(join(robot_path, "..", "tests")),
        ]

        test_directory = try_dirs[0]

        for d in try_dirs:
            if exists(d):
                test_directory = d
                break
        else:
            os.makedirs(test_directory)

        print("Tests directory is %s" % test_directory)
        print()
        builtin_tests_file = join(test_directory, "pyfrc_test.py")
        if exists(builtin_tests_file):
            print("- pyfrc_test.py already exists")
        else:
            with open(builtin_tests_file, "w") as fp:
                fp.write(builtin_tests)
            print("- builtin tests created at", builtin_tests_file)

        print()
        print("Robot tests can be ran via 'python3 robot.py test'")



from hal_impl.sim_hooks import SimHooks


class PyFrcFakeHooks(SimHooks):
    """
    Defines hal hooks that use the fake time object
    """

    def __init__(self, fake_time):
        self.fake_time = fake_time
        super().__init__()

    #
    # Time related hooks
    #

    def getTime(self):
        return self.fake_time.get()

    def delayMillis(self, ms):
        self.fake_time.increment_time_by(0.001 * ms)

    def delaySeconds(self, s):
        self.fake_time.increment_time_by(s)

    #
    # DriverStation related hooks
    #

    @property
    def ds_cond(self):
        return self.fake_time.ds_cond

    @ds_cond.setter
    def ds_cond(self, value):
        pass  # ignored



"""
    .. warning:: These drivetrain models are not particularly realistic, and
                 if you are using a tank drive style drivetrain you should use
                 the :class:`.TankModel` instead.

    Based on input from various drive motors, these helper functions
    simulate moving the robot in various ways. Many thanks to
    `Ether <http://www.chiefdelphi.com/forums/member.php?u=34863>`_
    for assistance with the motion equations.
      
    When specifying the robot speed to the below functions, the following
    may help you determine the approximate speed of your robot:
    
    * Slow: 4ft/s
    * Typical: 5 to 7ft/s
    * Fast: 8 to 12ft/s
        
    Obviously, to get the best simulation results, you should try to
    estimate the speed of your robot accurately.
    
    Here's an example usage of the drivetrains::
    
        import hal.simulation
        from pyfrc.physics import drivetrains
    
        class PhysicsEngine:
            
            def __init__(self, physics_controller):
                self.physics_controller = physics_controller
                self.drivetrain = drivetrains.TwoMotorDrivetrain(deadzone=drivetrains.linear_deadzone(0.2))
                
                self.l_motor = hal.simulation.PWMSim(1)
                self.r_motor = hal.simulation.PWMSim(2)
                
            def update_sim(self, now, tm_diff):
                l_motor = self.l_motor.getSpeed()
                r_motor = self.r_motor.getSpeed()

                speeds = self.drivetrain.calculate(l_motor, r_motor)
                self.physics_controller.drive(speeds, tm_diff)
                
                # optional: compute encoder
                # l_encoder = self.drivetrain.wheelSpeeds.left * tm_diff
    
    .. versionchanged:: 2020.1.0

       The input speeds and output rotation angles were changed to reflect
       the current WPILib drivetrain/field objects. Wheelbases and default
       speeds all require units.
"""
import math
import typing

from .units import units, Helpers

from wpilib.geometry import Translation2d

from wpilib.kinematics import (
    ChassisSpeeds,
    DifferentialDriveKinematics,
    DifferentialDriveWheelSpeeds,
    MecanumDriveKinematics,
    MecanumDriveWheelSpeeds,
)

DeadzoneCallable = typing.Callable[[float], float]


def linear_deadzone(deadzone: float) -> DeadzoneCallable:
    """
    Real motors won't actually move unless you give them some minimum amount
    of input. This computes an output speed for a motor and causes it to
    'not move' if the input isn't high enough. Additionally, the output is
    adjusted linearly to compensate.

    Example: For a deadzone of 0.2:

    * Input of 0.0 will result in 0.0
    * Input of 0.2 will result in 0.0
    * Input of 0.3 will result in ~0.12
    * Input of 1.0 will result in 1.0

    This returns a function that computes the deadzone. You should pass the
    returned function to one of the drivetrain simulation functions as the
    ``deadzone`` parameter.

    :param motor_input: The motor input (between -1 and 1)
    :param deadzone: Minimum input required for the motor to move (between 0 and 1)
    """
    assert 0.0 < deadzone < 1.0
    scale_param = 1.0 - deadzone

    def _linear_deadzone(motor_input):
        abs_motor_input = abs(motor_input)
        if abs_motor_input < deadzone:
            return 0.0
        else:
            return math.copysign(
                (abs_motor_input - deadzone) / scale_param, motor_input
            )

    return _linear_deadzone


class TwoMotorDrivetrain:
    """
    Two center-mounted motors with a simple drivetrain. The
    motion equations are as follows::

        FWD = (L+R)/2
        RCCW = (R-L)/W

    * L is forward speed of the left wheel(s), all in sync
    * R is forward speed of the right wheel(s), all in sync
    * W is wheelbase in feet

    .. note:: :class:`wpilib.drive.DifferentialDrive` assumes that to make
              the robot go forward, the left motor output is 1, and the
              right motor output is -1

    .. versionadded:: 2018.2.0
    """

    #: Wheel speeds you can use for encoder calculations (updated by calculate)
    wheelSpeeds: DifferentialDriveWheelSpeeds

    def __init__(
        self,
        x_wheelbase: units.Quantity = 2 * units.feet,
        speed: units.Quantity = 5 * units.fps,
        deadzone: typing.Optional[DeadzoneCallable] = None,
    ):
        """
        :param x_wheelbase: The distance between right and left wheels.
        :param speed:       Speed of robot (see above)
        :param deadzone:    A function that adjusts the output of the motor (see :func:`linear_deadzone`)
        """
        trackwidth = units.meters.m_from(x_wheelbase, name="x_wheelbase")
        self.kinematics = DifferentialDriveKinematics(trackwidth)
        self.speed = units.mps.m_from(speed, name="speed")
        self.wheelSpeeds = DifferentialDriveWheelSpeeds()
        self.deadzone = deadzone

    def calculate(self, l_motor: float, r_motor: float) -> ChassisSpeeds:
        """
        Given motor values, computes resulting chassis speeds of robot

        :param l_motor:    Left motor value (-1 to 1); 1 is forward
        :param r_motor:    Right motor value (-1 to 1); -1 is forward

        :returns: ChassisSpeeds that can be passed to 'drive'

        .. versionadded:: 2020.1.0
        """
        if self.deadzone:
            l_motor = self.deadzone(l_motor)
            r_motor = self.deadzone(r_motor)

        l = l_motor * self.speed
        r = -r_motor * self.speed

        self.wheelSpeeds.left = l
        self.wheelSpeeds.right = r

        return self.kinematics.toChassisSpeeds(self.wheelSpeeds)


class FourMotorDrivetrain:
    """
    Four motors, each side chained together. The motion equations are
    as follows::

        FWD = (L+R)/2
        RCCW = (R-L)/W

    * L is forward speed of the left wheel(s), all in sync
    * R is forward speed of the right wheel(s), all in sync
    * W is wheelbase in feet

    .. note:: :class:`wpilib.drive.DifferentialDrive` assumes that to make
              the robot go forward, the left motors must be set to 1, and
              the right to -1

    .. versionadded:: 2018.2.0
    """

    #: Wheel speeds you can use for encoder calculations (updated by calculate)
    wheelSpeeds: DifferentialDriveWheelSpeeds

    def __init__(
        self,
        x_wheelbase: units.Quantity = 2 * units.feet,
        speed: units.Quantity = 5 * units.fps,
        deadzone: typing.Optional[DeadzoneCallable] = None,
    ):
        """
        :param x_wheelbase: The distance between right and left wheels.
        :param speed:       Speed of robot (see above)
        :param deadzone:    A function that adjusts the output of the motor (see :func:`linear_deadzone`)
        """
        trackwidth = units.meters.m_from(x_wheelbase, name="x_wheelbase")
        self.kinematics = DifferentialDriveKinematics(trackwidth)
        self.speed = units.mps.m_from(speed, name="speed")
        self.wheelSpeeds = DifferentialDriveWheelSpeeds()
        self.deadzone = deadzone

    def calculate(
        self, lf_motor: float, lr_motor: float, rf_motor: float, rr_motor: float
    ) -> ChassisSpeeds:
        """
        Given motor values, computes resulting chassis speeds of robot

        :param lf_motor:   Left front motor value (-1 to 1); 1 is forward
        :param lr_motor:   Left rear motor value (-1 to 1); 1 is forward
        :param rf_motor:   Right front motor value (-1 to 1); -1 is forward
        :param rr_motor:   Right rear motor value (-1 to 1); -1 is forward

        :returns: ChassisSpeeds that can be passed to 'drive'

        .. versionadded:: 2020.1.0
        """

        if self.deadzone:
            lf_motor = self.deadzone(lf_motor)
            lr_motor = self.deadzone(lr_motor)
            rf_motor = self.deadzone(rf_motor)
            rr_motor = self.deadzone(rr_motor)

        l = (lf_motor + lr_motor) * 0.5 * self.speed
        r = -(rf_motor + rr_motor) * 0.5 * self.speed

        self.wheelSpeeds.left = l
        self.wheelSpeeds.right = r

        return self.kinematics.toChassisSpeeds(self.wheelSpeeds)


class MecanumDrivetrain:
    """
    Four motors, each with a mecanum wheel attached to it.

    .. note:: :class:`wpilib.drive.MecanumDrive` assumes that to make
              the robot go forward, the left motor outputs are 1, and the
              right motor outputs are -1

    .. versionadded:: 2018.2.0
    """

    #: Use this to compute encoder data after calculate is called
    wheelSpeeds: MecanumDriveWheelSpeeds

    def __init__(
        self,
        x_wheelbase: units.Quantity = 2 * units.feet,
        y_wheelbase: units.Quantity = 3 * units.feet,
        speed: units.Quantity = 5 * units.fps,
        deadzone: typing.Optional[DeadzoneCallable] = None,
    ):
        """
        :param x_wheelbase: The distance between right and left wheels.
        :param y_wheelbase: The distance between forward and rear wheels.
        :param speed:       Speed of robot (see above)
        :param deadzone:    A function that adjusts the output of the motor (see :func:`linear_deadzone`)
        """
        x2 = units.meters.m_from(x_wheelbase, name="x_wheelbase") / 2.0
        y2 = units.meters.m_from(y_wheelbase, name="y_wheelbase") / 2.0

        self.kinematics = MecanumDriveKinematics(
            Translation2d(x2, y2),
            Translation2d(x2, -y2),
            Translation2d(-x2, y2),
            Translation2d(-x2, -y2),
        )

        self.speed = units.mps.m_from(speed, name="speed")
        self.deadzone = deadzone

        self.wheelSpeeds = MecanumDriveWheelSpeeds()

    def calculate(
        self,
        lf_motor: float,
        lr_motor: float,
        rf_motor: float,
        rr_motor: float,
    ) -> ChassisSpeeds:
        """
        :param lf_motor:   Left front motor value (-1 to 1); 1 is forward
        :param lr_motor:   Left rear motor value (-1 to 1); 1 is forward
        :param rf_motor:   Right front motor value (-1 to 1); -1 is forward
        :param rr_motor:   Right rear motor value (-1 to 1); -1 is forward

        :returns: ChassisSpeeds that can be passed to 'drive'

        .. versionadded:: 2020.1.0
        """

        if self.deadzone:
            lf_motor = self.deadzone(lf_motor)
            lr_motor = self.deadzone(lr_motor)
            rf_motor = self.deadzone(rf_motor)
            rr_motor = self.deadzone(rr_motor)

        # Calculate speed of each wheel
        lr = lr_motor * self.speed
        rr = -rr_motor * self.speed
        lf = lf_motor * self.speed
        rf = -rf_motor * self.speed

        self.wheelSpeeds.frontLeft = lf
        self.wheelSpeeds.rearLeft = lr
        self.wheelSpeeds.frontRight = rf
        self.wheelSpeeds.rearRight = rr

        return self.kinematics.toChassisSpeeds(self.wheelSpeeds)


def four_motor_swerve_drivetrain(
    lr_motor: float,
    rr_motor: float,
    lf_motor: float,
    rf_motor: float,
    lr_angle: float,
    rr_angle: float,
    lf_angle: float,
    rf_angle: float,
    x_wheelbase=2,
    y_wheelbase=2,
    speed=5,
    deadzone=None,
) -> ChassisSpeeds:
    """
    Four motors that can be rotated in any direction

    If any motors are inverted, then you will need to multiply that motor's
    value by -1.

    :param lr_motor:   Left rear motor value (-1 to 1); 1 is forward
    :param rr_motor:   Right rear motor value (-1 to 1); 1 is forward
    :param lf_motor:   Left front motor value (-1 to 1); 1 is forward
    :param rf_motor:   Right front motor value (-1 to 1); 1 is forward

    :param lr_angle:   Left rear motor angle in degrees (0 to 360 measured clockwise from forward position)
    :param rr_angle:   Right rear motor angle in degrees (0 to 360 measured clockwise from forward position)
    :param lf_angle:   Left front motor angle in degrees (0 to 360 measured clockwise from forward position)
    :param rf_angle:   Right front motor angle in degrees (0 to 360 measured clockwise from forward position)

    :param x_wheelbase: The distance in feet between right and left wheels.
    :param y_wheelbase: The distance in feet between forward and rear wheels.
    :param speed:       Speed of robot in feet per second (see above)
    :param deadzone:    A function that adjusts the output of the motor (see :func:`linear_deadzone`)

    :returns: ChassisSpeeds that can be passed to 'drive'

    .. versionchanged:: 2020.1.0

       The output rotation angle was changed from CW to CCW to reflect the
       current WPILib drivetrain/field objects
    """

    if deadzone:
        lf_motor = deadzone(lf_motor)
        lr_motor = deadzone(lr_motor)
        rf_motor = deadzone(rf_motor)
        rr_motor = deadzone(rr_motor)

    # Calculate speed of each wheel
    lr = lr_motor * speed
    rr = rr_motor * speed
    lf = lf_motor * speed
    rf = rf_motor * speed

    # Calculate angle in radians
    lr_rad = math.radians(lr_angle)
    rr_rad = math.radians(rr_angle)
    lf_rad = math.radians(lf_angle)
    rf_rad = math.radians(rf_angle)

    # Calculate wheelbase radius
    wheelbase_radius = math.hypot(x_wheelbase / 2.0, y_wheelbase / 2.0)

    # Calculates the Vx and Vy components
    # Sin an Cos inverted because forward is 0 on swerve wheels
    Vx = (
        (math.sin(lr_rad) * lr)
        + (math.sin(rr_rad) * rr)
        + (math.sin(lf_rad) * lf)
        + (math.sin(rf_rad) * rf)
    )
    Vy = (
        (math.cos(lr_rad) * lr)
        + (math.cos(rr_rad) * rr)
        + (math.cos(lf_rad) * lf)
        + (math.cos(rf_rad) * rf)
    )

    # Adjusts the angle corresponding to a diameter that is perpendicular to the radius (add or subtract 45deg)
    lr_rad = (lr_rad + (math.pi / 4)) % (2 * math.pi)
    rr_rad = (rr_rad - (math.pi / 4)) % (2 * math.pi)
    lf_rad = (lf_rad - (math.pi / 4)) % (2 * math.pi)
    rf_rad = (rf_rad + (math.pi / 4)) % (2 * math.pi)

    # Finds the rotational velocity by finding the torque and adding them up
    Vw = wheelbase_radius * (
        (math.cos(lr_rad) * -lr)
        + (math.cos(rr_rad) * rr)
        + (math.cos(lf_rad) * -lf)
        + (math.cos(rf_rad) * rf)
    )

    Vx *= 0.25
    Vy *= 0.25
    Vw *= 0.25

    return ChassisSpeeds.fromFeet(Vx, Vy, Vw)



import argparse
import inspect
from os.path import abspath
import subprocess
import sys


class PyFrcProfiler:
    """
    Wraps other commands by running them via the built in cProfile module.
    Use this to profile your program and figure out where you're spending
    a lot of time (note that cProfile only profiles the main thread)
    """

    def __init__(self, parser):
        parser.add_argument(
            "-o", "--outfile", default=None, help="Save stats to <outfile>"
        )
        parser.add_argument(
            "args", nargs=argparse.REMAINDER, help="Arguments to pass to robot.py"
        )

    def run(self, options, robot_class, **static_options):

        print("profiling is not yet implemented for RobotPy 2020")
        return 1

        from .. import config

        config.mode = "profiler"

        try:
            import cProfile
        except ImportError:
            print(
                "Error importing cProfile module for profiling, your python interpreter may not support profiling\n",
                file=sys.stderr,
            )
            return 1

        if len(options.args) == 0:
            print("ERROR: Profiler command requires arguments to run other commands")
            return 1

        file_location = abspath(inspect.getfile(robot_class))

        if options.outfile:
            profile_args = ["-o", options.outfile]
        else:
            profile_args = ["-s", "tottime"]

        # construct the arguments to run the profiler
        args = (
            [sys.executable, "-m", "cProfile"]
            + profile_args
            + [file_location]
            + options.args
        )

        return subprocess.call(args)



from os.path import abspath, dirname
import argparse
import inspect
import logging
from pkg_resources import iter_entry_points

try:
    from importlib.metadata import metadata
except ImportError:
    from importlib_metadata import metadata

logger = logging.getLogger("pyfrc.sim")


class PyFrcSim:
    """
    Runs the robot using WPILib's GUI HAL Simulator
    """

    def __init__(self, parser: argparse.ArgumentParser):
        parser.add_argument(
            "--nogui",
            default=False,
            action="store_true",
            help="Don't use the WPIlib simulation gui",
        )

        self.simexts = {}

        for entry_point in iter_entry_points(group="robotpysimext", name=None):
            if entry_point.module_name == "halsim_gui":
                continue
            try:
                sim_ext_module = entry_point.load()
            except ImportError:
                print(f"WARNING: Error detected in {entry_point}")
                continue

            self.simexts[entry_point.name] = sim_ext_module

            parser.add_argument(
                f"--{entry_point.name}",
                default=False,
                action="store_true",
                help=metadata(entry_point.dist.project_name)["summary"],
            )

    def run(self, options, robot_class, **static_options):

        if not options.nogui:
            try:
                import halsim_gui
            except ImportError:
                print("robotpy-halsim-gui is not installed!")
                exit(1)
            else:
                halsim_gui.loadExtension()

        for name, module in self.simexts.items():
            if getattr(options, name.replace("-", "_"), False):
                try:
                    module.loadExtension()
                except:
                    print(f"Error loading {name}!")
                    raise

        # initialize physics, attach to the user robot class
        from ..physics.core import PhysicsInterface, PhysicsInitException

        robot_file = abspath(inspect.getfile(robot_class))
        robot_path = dirname(robot_file)

        try:
            physics = PhysicsInterface._create(robot_path)
            if physics:
                robot_class._simulationInit = physics._simulationInit
                robot_class._simulationPeriodic = physics._simulationPeriodic

            # run the robot
            return robot_class.main(robot_class)

        except PhysicsInitException:
            return False



"""
    pyfrc supports simplistic custom physics model implementations for
    simulation and testing support. It can be as simple or complex as you want
    to make it. We will continue to add helper functions (such as the
    :mod:`pyfrc.physics.drivetrains` module) to make this a lot easier
    to do. General purpose physics implementations are welcome also!

    The idea is you provide a :class:`PhysicsEngine` object that interacts with
    the simulated HAL, and modifies motors/sensors accordingly depending on the
    state of the simulation. An example of this would be measuring a motor
    moving for a set period of time, and then changing a limit switch to turn
    on after that period of time. This can help you do more complex simulations
    of your robot code without too much extra effort.

    By default, pyfrc doesn't modify any of your inputs/outputs without being
    told to do so by your code or the simulation GUI.
    
    See the `physics sample <https://github.com/robotpy/examples/tree/master/physics/src>`_
    for more details.

    Enabling physics support
    ------------------------

    You must create a python module called ``physics.py`` next to your
    ``robot.py``. A physics module must have a class called
    :class:`PhysicsEngine` which must have a function called ``update_sim``.
    When initialized, it will be passed an instance of this object.
"""

import imp
import math
from os.path import exists, join
import threading

import wpilib
import wpilib.simulation

from wpimath.kinematics import ChassisSpeeds
from wpimath.geometry import Pose2d, Rotation2d, Transform2d, Translation2d, Twist2d

import logging

logger = logging.getLogger("pyfrc.physics")


class PhysicsInitException(Exception):
    pass


class PhysicsEngine:
    """
    Your physics module must contain a class called ``PhysicsEngine``,
    and it must implement the same functions as this class.

    Alternatively, you can inherit from this object. However, that is
    not required.
    """

    def __init__(self, physics_controller: "PhysicsInterface"):
        """
        The constructor must take the following arguments:

        :param physics_controller: The physics controller interface
        :type  physics_controller: :class:`.PhysicsInterface`
        """
        self.physics_controller = physics_controller

    def update_sim(self, now: float, tm_diff: float):
        """
        Called when the simulation parameters for the program should be
        updated. This is called after robotPeriodic is called.

        :param now: The current time
        :type  now: float
        :param tm_diff: The amount of time that has passed since the last
                        time that this function was called
        :type  tm_diff: float
        """
        pass


class PhysicsInterface:
    """
    An instance of this is passed to the constructor of your
    :class:`PhysicsEngine` object. This instance is used to communicate
    information to the simulation, such as moving the robot on the
    field displayed to the user.
    """

    @classmethod
    def _create(cls, robot_path):

        physics_module_path = join(robot_path, "physics.py")
        if exists(physics_module_path):

            # Load the user's physics module if it exists
            try:
                physics_module = imp.load_source("physics", physics_module_path)
            except:
                logger.exception("Error loading user physics module")
                raise PhysicsInitException()

            if not hasattr(physics_module, "PhysicsEngine"):
                logger.error("User physics module does not have a PhysicsEngine object")
                raise PhysicsInitException()

            logger.info("Physics support successfully enabled")
            return PhysicsInterface(physics_module)

        else:
            logger.warning(
                "Cannot enable physics support, %s not found", physics_module_path
            )

    def __init__(self, physics_module):
        self.last_tm = None
        self.module = physics_module
        self.engine = None
        self.device_gyro_channels = []
        self.field = wpilib.Field2d()
        wpilib.SmartDashboard.putData("Field", self.field)

    def __repr__(self):
        return "Physics"

    def _simulationInit(self):
        # look for a class called PhysicsEngine
        try:
            self.engine = self.module.PhysicsEngine(self)
        except Exception:
            logger.exception("Error creating user's PhysicsEngine object")
            raise PhysicsInitException()

    def _simulationPeriodic(self):
        now = wpilib.Timer.getFPGATimestamp()
        last_tm = self.last_tm

        if last_tm is None:
            self.last_tm = now
        else:

            # When using time, always do it based on a differential! You may
            # not always be called at a constant rate
            tm_diff = now - last_tm

            # Don't run physics calculations more than 100hz
            if tm_diff > 0.010:
                try:
                    self.engine.update_sim(now, tm_diff)
                except Exception as e:
                    raise Exception(
                        "User physics code raised an exception (see above)"
                    ) from e
                self.last_tm = now

    #######################################################
    #
    # Public API
    #
    #######################################################

    def drive(self, speeds: ChassisSpeeds, tm_diff: float) -> Pose2d:
        """Call this from your :func:`PhysicsEngine.update_sim` function.
        Will update the robot's position on the simulation field.

        You can either calculate the chassis speeds yourself, or you
        can use the predefined functions in :mod:`pyfrc.physics.drivetrains`.

        The outputs of the `drivetrains.*` functions should be passed
        to this function.

        :param speeds:   Represents current speed/angle of robot travel
        :param tm_diff:  Amount of time speed was traveled (this is the
                         same value that was passed to update_sim)

        :return: current robot pose

        .. versionchanged:: 2020.1.0
           Input parameter is ChassisSpeeds object
        """

        twist = Twist2d(
            dx=speeds.vx * tm_diff,
            dy=speeds.vy * tm_diff,
            dtheta=speeds.omega * tm_diff,
        )

        pose = self.field.getRobotPose()
        pose = pose.exp(twist)
        self.field.setRobotPose(pose)
        return pose

    def move_robot(self, transform: Transform2d) -> Pose2d:
        """Call this from your :func:`PhysicsEngine.update_sim` function.
        Will update the robot's position on the simulation field.

        This moves the robot some relative distance and angle from
        its current position.

        :param transform: The distance and angle to move the robot

        :return: current robot pose

        .. versionadded:: 2020.1.0
        """

        pose = self.field.getRobotPose()
        pose = pose + transform
        self.field.setRobotPose(pose)
        return pose

    def get_pose(self):
        """
        :returns: current robot pose

        .. versionadded:: 2020.1.0
        """
        return self.field.getRobotPose()

    # def get_offset(self, point: Translation2d):
    #     """
    #         Computes how far away and at what angle a coordinate is
    #         located from the robot.

    #         :returns: distance,angle offset of the given x,y coordinate

    #         .. versionadded:: 2018.1.7

    #         .. versionchanged:: 2020.1.0
    #     """
    #     tr = self.field.getRobotPose().translation()

    #     dx = tr.x_feet - x
    #     dy = tr.y_feet - y

    #     distance = math.hypot(dx, dy)
    #     angle = math.atan2(dy, dx)
    #     return distance, math.degrees(angle)



import pytest
from pyfrc.physics import drivetrains
from pyfrc.physics.units import units
from math import sqrt


@pytest.mark.parametrize(
    "l_motor,r_motor,output",
    [
        (0, 0, (0, 0)),  # stationary
        (1, -1, (1, 0)),  # forward
        (-1, 1, (-1, 0)),  # backwards
        (-1, -1, (0, 1)),  # rotate left
        (1, 1, (0, -1)),  # rotate right
    ],
)
def test_two_motor_drivetrain(l_motor, r_motor, output):
    result = drivetrains.TwoMotorDrivetrain(speed=1 * units.fps).calculate(
        l_motor, r_motor
    )
    result = (result.vx_fps, result.omega)
    assert abs(result[0] - output[0]) < 0.001
    assert abs(result[1] - output[1]) < 0.001


@pytest.mark.parametrize(
    "lf_motor,lr_motor,rf_motor,rr_motor,output",
    [
        (0, 0, 0, 0, (0, 0)),  # stationary
        (1, 1, -1, -1, (1, 0)),  # forward
        (-1, -1, 1, 1, (-1, 0)),  # backwards
        (-1, -1, -1, -1, (0, 1)),  # rotate left
        (1, 1, 1, 1, (0, -1)),  # rotate right
    ],
)
def test_four_motor_drivetrain(lf_motor, lr_motor, rf_motor, rr_motor, output):
    result = drivetrains.FourMotorDrivetrain(speed=1 * units.fps).calculate(
        lf_motor, lr_motor, rf_motor, rr_motor
    )
    result = (result.vx_fps, result.omega)
    assert abs(result[0] - output[0]) < 0.001
    assert abs(result[1] - output[1]) < 0.001


# @pytest.mark.parametrize(
#     "lr_motor,rr_motor,lf_motor,rf_motor,output",
#     [
#         (0, 0, 0, 0, (0, 0, 0)),  # stationary
#         (1, 1, 1, 1, (0, 1, 0)),  # forward
#         (-1, -1, -1, -1, (0, -1, 0)),  # backwards
#         (1, -1, -1, 1, (-1, 0, 0)),  # strafe left
#         (-1, 1, 1, -1, (1, 0, 0)),  # strafe right
#         (-1, 1, -1, 1, (0, 0, 1)),  # rotate left
#         (1, -1, 1, -1, (0, 0, -1)),  # rotate right
#     ],
# )
# def test_mecanum_drivetrain(lr_motor, rr_motor, lf_motor, rf_motor, output):
#     result = drivetrains.MecanumDrivetrain(
#         x_wheelbase=1, y_wheelbase=1, speed=1,
#     ).get_vector(lf_motor, lr_motor, rf_motor, rr_motor)
#     assert abs(result[0] - output[0]) < 0.001
#     assert abs(result[1] - output[1]) < 0.001
#     assert abs(result[2] - output[2]) < 0.001


@pytest.mark.parametrize(
    "lr_motor,rr_motor,lf_motor,rf_motor,lr_angle,rr_angle,lf_angle,rf_angle,output",
    [
        (0, 0, 0, 0, 0, 0, 0, 0, (0, 0, 0)),  # stationary
        (1, 1, 1, 1, 0, 0, 0, 0, (0, 1, 0)),  # forward
        (-1, -1, -1, -1, 180, 180, 180, 180, (0, 1, 0)),  # forward inverted
        (-1, -1, -1, -1, 0, 0, 0, 0, (0, -1, 0)),  # backward
        (1, 1, 1, 1, 180, 180, 180, 180, (0, -1, 0)),  # backward inverted
        (1, 1, 1, 1, 270, 270, 270, 270, (-1, 0, 0)),  # strafe left
        (-1, -1, -1, -1, 90, 90, 90, 90, (-1, 0, 0)),  # strafe left inverted
        (1, 1, 1, 1, 90, 90, 90, 90, (1, 0, 0)),  # strafe right
        (-1, -1, -1, -1, 270, 270, 270, 270, (1, 0, 0)),  # strafe right inverted
        (1, 1, 1, 1, 135, 45, 225, 315, (0, 0, 1)),  # rotate left
        (-1, -1, -1, -1, 315, 225, 45, 135, (0, 0, 1)),  # rotate left inverted
        (1, 1, 1, 1, 315, 225, 45, 135, (0, 0, -1)),  # rotate right
        (-1, -1, -1, -1, 135, 45, 225, 315, (0, 0, -1)),  # rotate right inverted
    ],
)
def test_four_motor_swerve_drivetrain(
    lr_motor,
    rr_motor,
    lf_motor,
    rf_motor,
    lr_angle,
    rr_angle,
    lf_angle,
    rf_angle,
    output,
):
    wheelbase = sqrt(2)
    result = drivetrains.four_motor_swerve_drivetrain(
        lr_motor,
        rr_motor,
        lf_motor,
        rf_motor,
        lr_angle,
        rr_angle,
        lf_angle,
        rf_angle,
        speed=1,
        x_wheelbase=wheelbase,
        y_wheelbase=wheelbase,
    )

    assert abs(result.vx_fps - output[0]) < 0.001
    assert abs(result.vy_fps - output[1]) < 0.001
    assert abs(result.omega - output[2]) < 0.001



import argparse
import inspect

from os.path import abspath, dirname, join

from robotpy_installer import sshcontroller

from ..util import print_err, yesno


class PyFrcUndeploy:
    """
    Removes current robot code from a RoboRIO
    """

    def __init__(self, parser: argparse.ArgumentParser):

        robot_args = parser.add_mutually_exclusive_group()

        robot_args.add_argument(
            "--robot", default=None, help="Set hostname or IP address of robot"
        )

        robot_args.add_argument(
            "--team", default=None, type=int, help="Set team number to deploy robot for"
        )

        parser.add_argument(
            "--no-resolve",
            action="store_true",
            default=False,
            help="If specified, don't do a DNS lookup, allow ssh et al to do it instead",
        )

    def run(self, options, robot_class, **static_options):
        from .. import config

        config.mode = "undeploy"

        robot_file = abspath(inspect.getfile(robot_class))
        robot_path = dirname(robot_file)
        cfg_filename = join(robot_path, ".deploy_cfg")

        if not yesno(
            "This will stop your robot code and delete it from the RoboRIO. Continue?"
        ):
            return 1

        hostname_or_team = options.robot
        if not hostname_or_team and options.team:
            hostname_or_team = options.team

        try:
            with sshcontroller.ssh_from_cfg(
                cfg_filename,
                username="lvuser",
                password="",
                hostname=hostname_or_team,
                no_resolve=options.no_resolve,
            ) as ssh:

                # first, turn off the running program
                ssh.exec_cmd("/usr/local/frc/bin/frcKillRobot.sh -t")

                # delete the code
                ssh.exec_cmd("rm -rf /home/lvuser/py")

                # for good measure, delete the start command too
                ssh.exec_cmd(
                    "rm -f /home/lvuser/robotDebugCommand /home/lvuser/robotCommand"
                )

        except sshcontroller.SshExecError as e:
            print_err("ERROR:", str(e))
            return 1

        print("SUCCESS: Files have been successfully wiped!")

        return 0



"""
    The purpose of the fuzz 'test' is not exactly to 'do' anything, but
    rather it mashes the buttons and switches in various completely
    random ways to try and find any possible control situations and
    such that would probably never *normally* come up, but.. well,
    given a bit of bad luck, could totally happen.
    
    Keep in mind that the results will totally different every time
    you run this, so if you find an error, fix it -- but don't expect
    that you'll be able to duplicate it with this test. Instead, you
    should design a specific test that can trigger the bug, to ensure
    that you actually fixed it.
"""

import random
import math

from .. import config

_gsms = config.config_obj["pyfrc"]["game_specific_messages"]


def fuzz_bool():
    return random.randrange(0, 2, 1) == 0


def fuzz_all(hal_data):

    # fuzz the eio switches
    for dio in hal_data["dio"]:

        # inputs only
        if not dio["is_input"] or not dio["initialized"]:
            continue

        # activate at random times
        dio["value"] = fuzz_bool()

    # fuzz the joysticks
    for stick in hal_data["joysticks"]:
        # if stick['has_source']:
        # axes
        for axis in range(len(stick["axes"])):
            if fuzz_bool():
                stick["axes"][axis] = random.uniform(-1, 1)

        # buttons
        for button in range(len(stick["buttons"])):
            stick["buttons"][button] = fuzz_bool()

    # fuzz analog channels
    for analog in hal_data["analog_in"]:
        # if analog['has_source'] and fuzz_bool():
        analog["voltage"] = analog["avg_voltage"] = random.uniform(0.0, 5.0)
        analog["value"] = int(analog["voltage"] / 5.0 * analog["offset"])


def test_fuzz(hal_data, control, fake_time, robot):
    """
    Runs through a whole game randomly setting components
    """

    class TestController:
        def __init__(self):
            self.mode = None

            self.disabled = 0
            self.autonomous = 0
            self.teleop = 0

        def on_step(self, tm):
            """
            Called on each simulation step. This runs through each mode,
            and asserts that the robot didn't spend too much time in any
            particular mode.This also calls fuzz_all and fuzzes all data

            :param tm: The current robot time
            """

            fuzz_all(hal_data)
            mode = control.get_mode()
            if mode == self.mode:
                return

            if mode == "autonomous":
                self.autonomous += 1
                assert int(math.floor(fake_time.get())) == 5

            elif mode == "teleop":
                self.teleop += 1
                assert int(math.floor(fake_time.get())) == 21

            elif mode == "disabled":
                self.disabled += 1

                if self.disabled == 1:
                    assert int(math.floor(fake_time.get())) == 0
                else:
                    assert int(math.floor(fake_time.get())) == 20
            else:
                assert False, "Internal error!"

            self.mode = mode

    if _gsms:
        control.game_specific_message = _gsms[0]

    control.set_practice_match()
    tc = control.run_test(TestController)

    assert int(math.floor(fake_time.get())) == 36

    # If an error occurs here, for some reason a mode got called too many times
    assert tc.disabled == 2
    assert tc.autonomous == 1
    assert tc.teleop == 1



import os
import inspect
import sys

from os.path import abspath, dirname, exists, join

import pytest

from ..util import yesno

# from ..test_support import pytest_plugin

# TODO: setting the plugins so that the end user can invoke py.test directly
# could be a useful thing. Will have to consider that later.


class _TryAgain(Exception):
    pass


#
# main test class
#
class PyFrcTest:
    """
    Executes unit tests on the robot code using a special py.test plugin
    """

    def __init__(self, parser=None):
        if parser:
            parser.add_argument(
                "--builtin",
                default=False,
                action="store_true",
                help="Use pyfrc's builtin tests if no tests are specified",
            )
            parser.add_argument(
                "--coverage-mode",
                default=False,
                action="store_true",
                help="This flag is passed when trying to determine coverage",
            )
            parser.add_argument(
                "pytest_args",
                nargs="*",
                help="To pass args to pytest, specify --<space>, then the args",
            )

    def run(self, options, robot_class, **static_options):
        # wrapper around run_test that sets the appropriate mode

        print("tests are not yet implemented for RobotPy 2020")
        return 1

        from .. import config

        config.mode = "test"
        config.coverage_mode = options.coverage_mode

        return self.run_test(
            options.pytest_args, robot_class, options.builtin, **static_options
        )

    def run_test(self, *a, **k):
        try:
            return self._run_test(*a, **k)
        except _TryAgain:
            return self._run_test(*a, **k)

    def _run_test(self, pytest_args, robot_class, use_builtin, **static_options):

        # find test directory, change current directory so py.test can find the tests
        # -> assume that tests reside in tests or ../tests

        curdir = abspath(os.getcwd())

        self.robot_class = robot_class
        robot_file = abspath(inspect.getfile(robot_class))
        robot_path = dirname(robot_file)

        if robot_file.endswith("cProfile.py"):
            # so, the module for the robot class is __main__, and __main__ is
            # cProfile so try to find it
            robot_file = join(curdir, "robot.py")
            robot_path = curdir

            if not exists(robot_file):
                print(
                    "ERROR: Cannot run profiling from a directory that does not contain robot.py"
                )
                return 1

        self.try_dirs = [
            abspath(join(robot_path, "tests")),
            abspath(join(robot_path, "..", "tests")),
        ]

        for d in self.try_dirs:
            if exists(d):
                test_directory = d
                os.chdir(test_directory)
                break
        else:
            if not use_builtin:
                print("ERROR: Cannot run robot tests, as test directory was not found!")
                retv = self._no_tests()
                return 1

            from ..tests import basic

            pytest_args.insert(0, abspath(inspect.getfile(basic)))

        try:
            retv = pytest.main(
                pytest_args,
                plugins=[
                    pytest_plugin.PyFrcPlugin(robot_class, robot_file, robot_path)
                ],
            )
        finally:
            os.chdir(curdir)

        # requires pytest 2.8.x
        if retv == 5:
            print()
            print("ERROR: a tests directory was found, but no tests were defined")
            retv = self._no_tests(retv)

        return retv

    def _no_tests(self, r=1):
        print()
        print("Looked for tests at:")
        for d in self.try_dirs:
            print("- %s" % d)
        print()
        print(
            "If you don't want to write your own tests, pyfrc comes with generic tests"
        )
        print("that can test basic functionality of most robots. You can run them by")
        print("specifying the --builtin option.")
        print()

        if not sys.stdin.isatty():
            print(
                "Alternatively, to create a tests directory with the builtin tests, you can run:"
            )
            print()
            print("    python3 robot.py add-tests")
            print()
        else:
            if yesno("Create a tests directory with builtin tests now?"):
                from .cli_add_tests import PyFrcAddTests

                add_tests = PyFrcAddTests()
                add_tests.run(None, self.robot_class)

                raise _TryAgain()

        return r



import hal
import typing


class LinearMotion:
    """
    Helper for simulating motion involving a encoder directly coupled to
    a motor.

    Here's an example that shows a linear motion of 6ft at 2 ft/s with a
    360-count-per-ft encoder, coupled to a PWM motor on port 0 and the
    first encoder object::

        import hal.simulation
        from pyfrc.physics import motion

        class PhysicsEngine:

            def __init__(self, physics_controller):
                self.motion = pyfrc.physics.motion.LinearMotion('Motion', 2, 360, 6)
                self.motor = hal.simulation.PWMSim(0)
                self.encoder = hal.simulation

            def update_sim(self, now, tm_diff):
                motor_value = self.motor.getValue()
                count = self.motion.compute(motor_value, tm_diff)
                self.encoder.setCount(count)

    .. versionadded:: 2018.3.0
    """

    #: Current computed position of motion, in feet
    position_ft = 0

    #: Current computed position of motion (encoder ticks)
    position_ticks = 0

    def __init__(
        self,
        name: str,
        motor_ft_per_sec: float,
        ticks_per_feet: int,
        max_position: typing.Optional[float] = None,
        min_position: typing.Optional[float] = 0,
    ):
        """
        :param name: Name of motion, shown in simulation UI
        :param motor_ft_per_sec: Motor travel in feet per second (or whatever units you want)
        :param ticks_per_feet: Number of encoder ticks per feet
        :param max_position: Maximum position that this motion travels to
        :param min_position: Minimum position that this motion travels to
        """
        self.name = name
        self.motor_ft_per_sec = motor_ft_per_sec
        self.ticks_per_feet = ticks_per_feet
        self.min_position = min_position
        self.max_position = max_position

        self.device = hal.SimDevice(self.name)
        self.position = self.device.createDouble("position", True, 0.0)

    def compute(self, motor_val, tm_diff):
        self.position_ft += motor_val * tm_diff * self.motor_ft_per_sec

        if self.max_position is not None:
            self.position_ft = min(
                max(self.position_ft, self.min_position), self.max_position
            )

        self.position_ticks = int(self.position_ft * self.ticks_per_feet)

        # This causes a label to be rendered in the UI
        self.position.set(self.position_ft)
        return self.position_ticks



import argparse
import inspect
from os.path import dirname
import subprocess
import sys


class PyFrcCoverage:
    """
    Wraps other commands by running them via the coverage module. Requires
    the coverage module to be installed.
    """

    def __init__(self, parser):
        parser.add_argument(
            "args", nargs=argparse.REMAINDER, help="Arguments to pass to robot.py"
        )

    def run(self, options, robot_class, **static_options):

        print("tests+coverage are not yet implemented for RobotPy 2020")
        return 1

        try:
            import coverage
        except ImportError:
            print(
                "Error importing coverage module for code coverage testing, did you install it?\n"
                + "You can download it at https://pypi.python.org/pypi/coverage\n",
                file=sys.stderr,
            )
            return 1

        if len(options.args) == 0:
            print("ERROR: Coverage command requires arguments to run other commands")
            return 1

        file_location = inspect.getfile(robot_class)

        option_args = list(options.args)
        if option_args[0] == "test":
            option_args.insert(1, "--coverage-mode")

        # construct the arguments to run coverage
        args = [
            sys.executable,
            "-m",
            "coverage",
            "run",
            "--source",
            dirname(file_location),
            file_location,
        ] + list(options.args)

        retval = subprocess.call(args)
        if retval != 0:
            return retval

        args = [sys.executable, "-m", "coverage", "report", "-m"]

        return subprocess.call(args)
